#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .run_script import execute_with_odps_context
from .filesystem import VolumeFileSystem

from mars import __version__ as mars_version
from mars.lib.filesystem.core import register_filesystem
from mars.remote.run_script import RunScript
from mars.remote.core import RemoteFunction


RemoteFunction.register_executor(execute_with_odps_context(RemoteFunction.execute))
RunScript.register_executor(execute_with_odps_context(RunScript.execute))
register_filesystem("odps", VolumeFileSystem)


# hotfix v0.8 bug
if mars_version.startswith("0.8"):
    import pandas as pd
    from mars import utils as mars_utils

    def on_serialize_nsplits(value):
        if value is None:
            return None
        new_nsplits = []
        for dim_splits in value:
            new_nsplits.append(tuple(None if pd.isna(v) else v for v in dim_splits))
        return tuple(new_nsplits)

    mars_utils.on_serialize_nsplits.__code__ = on_serialize_nsplits.__code__

try:
    from pandas._libs import lib as _pandas__libs_lib
    from mars import utils as mars_utils

    if not hasattr(_pandas__libs_lib, "NoDefault"):
        _pandas__libs_lib.NoDefault = mars_utils.NoDefault
        _pandas__libs_lib.no_default = mars_utils.no_default
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import uuid
import logging

from mars.dataframe.operands import DataFrameOperandMixin, DataFrameOperand
from mars.dataframe.utils import parse_index
from mars.serialize import (
    StringField,
    SeriesField,
    BoolField,
    DictField,
    Int64Field,
    ListField,
    ValueType,
)
from mars.dataframe.utils import build_concatenated_rows_frame

from ....utils import to_str
from ...utils import convert_pandas_object_to_string

logger = logging.getLogger(__name__)

try:
    from mars.core import OutputType

    _output_type_kw = dict(_output_types=[OutputType.dataframe])
except ImportError:
    from mars.dataframe.operands import ObjectType

    _output_type_kw = dict(_object_type=ObjectType.dataframe)


class DataFrameWriteTable(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123460

    _dtypes = SeriesField("dtypes")

    _odps_params = DictField("odps_params")
    _table_name = StringField("table_name")
    _partition_spec = StringField("partition_spec")
    _partition_columns = ListField("partition_columns", ValueType.string)
    _overwrite = BoolField("overwrite")
    _write_batch_size = Int64Field("write_batch_size")
    _unknown_as_string = BoolField("unknown_as_string")

    def __init__(
        self,
        dtypes=None,
        odps_params=None,
        table_name=None,
        partition_spec=None,
        unknown_as_string=None,
        overwrite=None,
        write_batch_size=None,
        **kw
    ):
        kw.update(_output_type_kw)
        super(DataFrameWriteTable, self).__init__(
            _dtypes=dtypes,
            _odps_params=odps_params,
            _table_name=table_name,
            _partition_spec=partition_spec,
            _unknown_as_string=unknown_as_string,
            _overwrite=overwrite,
            _write_batch_size=write_batch_size,
            **kw
        )

    @property
    def retryable(self):
        return False

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def unknown_as_string(self):
        return self._unknown_as_string

    @property
    def odps_params(self):
        return self._odps_params

    @property
    def table_name(self):
        return self._table_name

    @property
    def partition_spec(self):
        return self._partition_spec

    @property
    def overwrite(self):
        return self._overwrite

    @property
    def write_batch_size(self):
        return self._write_batch_size

    @property
    def partition_columns(self):
        return self._partition_columns

    def __call__(self, x):
        shape = (0,) * len(x.shape)
        index_value = parse_index(x.index_value.to_pandas()[:0], x.key, "index")
        columns_value = parse_index(
            x.columns_value.to_pandas()[:0], x.key, "columns", store_data=True
        )
        return self.new_dataframe(
            [x],
            shape=shape,
            dtypes=x.dtypes[:0],
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def _tile_cupid(cls, op):
        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.runtime import RuntimeContext

        if not RuntimeContext.is_context_ready():
            raise SystemError(
                "No Mars cluster found, please create via `o.create_mars_cluster`."
            )
        cupid_ctx = context()

        bearer_token = cupid_ctx.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            None,
            None,
            account=account,
            project=odps_params["project"],
            endpoint=endpoint,
        )
        cupid_session = CupidSession(o)

        data_src = o.get_table(op.table_name)

        logger.debug("Start creating upload session from cupid.")
        upload_session = cupid_session.create_upload_session(data_src)

        input_df = build_concatenated_rows_frame(op.inputs[0])
        out_df = op.outputs[0]

        out_chunks = []
        out_chunk_shape = (0,) * len(input_df.shape)
        blocks = {}
        for chunk in input_df.chunks:
            block_id = str(int(time.time())) + "_" + str(uuid.uuid4()).replace("-", "")
            chunk_op = DataFrameWriteTableSplit(
                dtypes=op.dtypes,
                table_name=op.table_name,
                unknown_as_string=op.unknown_as_string,
                partition_spec=op.partition_spec,
                cupid_handle=to_str(upload_session.handle),
                block_id=block_id,
                write_batch_size=op.write_batch_size,
            )
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=out_chunk_shape,
                index=chunk.index,
                index_value=out_df.index_value,
                dtypes=chunk.dtypes,
            )
            out_chunks.append(out_chunk)
            blocks[block_id] = op.partition_spec

        # build commit tree
        combine_size = 8
        chunks = out_chunks
        while len(chunks) >= combine_size:
            new_chunks = []
            for i in range(0, len(chunks), combine_size):
                chks = chunks[i : i + combine_size]
                if len(chks) == 1:
                    chk = chks[0]
                else:
                    chk_op = DataFrameWriteTableCommit(
                        dtypes=op.dtypes, is_terminal=False
                    )
                    chk = chk_op.new_chunk(
                        chks,
                        shape=out_chunk_shape,
                        index_value=out_df.index_value,
                        dtypes=op.dtypes,
                    )
                new_chunks.append(chk)
            chunks = new_chunks

        assert len(chunks) < combine_size

        commit_table_op = DataFrameWriteTableCommit(
            dtypes=op.dtypes,
            table_name=op.table_name,
            blocks=blocks,
            cupid_handle=to_str(upload_session.handle),
            overwrite=op.overwrite,
            odps_params=op.odps_params,
            is_terminal=True,
        )
        commit_table_chunk = commit_table_op.new_chunk(
            chunks,
            shape=out_chunk_shape,
            dtypes=op.dtypes,
            index_value=out_df.index_value,
        )

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            shape=out_df.shape,
            index_value=out_df.index_value,
            dtypes=out_df.dtypes,
            columns_value=out_df.columns_value,
            chunks=[commit_table_chunk],
            nsplits=((0,),) * len(out_chunk_shape),
        )

    @classmethod
    def _tile_tunnel(cls, op):
        from odps import ODPS

        out_df = op.outputs[0]
        in_df = build_concatenated_rows_frame(op.inputs[0])

        if op.overwrite:
            o = ODPS(
                op.odps_params["access_id"],
                op.odps_params["secret_access_key"],
                project=op.odps_params["project"],
                endpoint=op.odps_params["endpoint"],
            )
            data_target = o.get_table(op.table_name)
            if op.partition_spec:
                data_target = data_target.get_partition(op.partition_spec)
            data_target.truncate()

        out_chunks = []
        for chunk in in_df.chunks:
            chunk_op = DataFrameWriteTableSplit(
                dtypes=op.dtypes,
                table_name=op.table_name,
                odps_params=op.odps_params,
                partition_spec=op.partition_spec,
            )
            index_value = parse_index(chunk.index_value.to_pandas()[:0], chunk)
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=(0, 0),
                index_value=index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
                index=chunk.index,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_df.params.copy()
        params.update(
            dict(chunks=out_chunks, nsplits=((0,) * in_df.chunk_shape[0], (0,)))
        )
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def tile(cls, op):
        from cupid.runtime import RuntimeContext

        if RuntimeContext.is_context_ready():
            return cls._tile_cupid(op)
        else:
            return cls._tile_tunnel(op)


class DataFrameWriteTableSplit(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123461

    _dtypes = SeriesField("dtypes")

    _table_name = StringField("table_name")
    _partition_spec = StringField("partition_spec")
    _cupid_handle = StringField("cupid_handle")
    _block_id = StringField("block_id")
    _write_batch_size = Int64Field("write_batch_size")
    _unknown_as_string = BoolField("unknown_as_string")

    # for tunnel
    _odps_params = DictField("odps_params")

    def __init__(
        self,
        dtypes=None,
        table_name=None,
        odps_params=None,
        partition_spec=None,
        cupid_handle=None,
        unknown_as_string=None,
        block_id=None,
        write_batch_size=None,
        **kw
    ):
        kw.update(_output_type_kw)
        super(DataFrameWriteTableSplit, self).__init__(
            _dtypes=dtypes,
            _table_name=table_name,
            _odps_params=odps_params,
            _partition_spec=partition_spec,
            _unknown_as_string=unknown_as_string,
            _cupid_handle=cupid_handle,
            _block_id=block_id,
            _write_batch_size=write_batch_size,
            **kw
        )

    @property
    def retryable(self):
        return False

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def table_name(self):
        return self._table_name

    @property
    def odps_params(self):
        return self._odps_params

    @property
    def unknown_as_string(self):
        return self._unknown_as_string

    @property
    def partition_spec(self):
        return self._partition_spec

    @property
    def cupid_handle(self):
        return self._cupid_handle

    @property
    def block_id(self):
        return self._block_id

    @property
    def write_batch_size(self):
        return self._write_batch_size

    @classmethod
    def _execute_in_cupid(cls, ctx, op):
        import pyarrow as pa
        import pandas as pd
        from ....df.backends.pd.types import pd_to_df_schema
        from cupid.io.table.core import BlockWriter

        to_store_data = ctx[op.inputs[0].key]

        odps_schema = pd_to_df_schema(
            to_store_data, unknown_as_string=op.unknown_as_string
        )
        project_name, table_name = op.table_name.split(".")
        if table_name.startswith('`') and table_name.endswith('`'):
            table_name = table_name[1:-1]
        block_writer = BlockWriter(
            _table_name=table_name,
            _project_name=project_name,
            _table_schema=odps_schema,
            _partition_spec=op.partition_spec,
            _block_id=op.block_id,
            _handle=op.cupid_handle,
        )
        logger.debug("Start writing table block, block id: %s", op.block_id)
        with block_writer.open_arrow_writer() as cupid_writer:

            sink = pa.BufferOutputStream()

            batch_size = op.write_batch_size or 1024
            batch_idx = 0
            batch_data = to_store_data[
                batch_size * batch_idx : batch_size * (batch_idx + 1)
            ]
            batch_data = convert_pandas_object_to_string(batch_data)
            schema = pa.RecordBatch.from_pandas(
                to_store_data[:1], preserve_index=False
            ).schema
            arrow_writer = pa.RecordBatchStreamWriter(sink, schema)
            while len(batch_data) > 0:
                batch = pa.RecordBatch.from_pandas(batch_data, preserve_index=False)
                arrow_writer.write_batch(batch)
                batch_idx += 1
                batch_data = to_store_data[
                    batch_size * batch_idx : batch_size * (batch_idx + 1)
                ]
            arrow_writer.close()
            cupid_writer.write(sink.getvalue())
        logger.debug("Write table block finished, block id: %s", op.block_id)

        block_writer.commit()
        ctx[op.outputs[0].key] = pd.DataFrame()

    @classmethod
    def _execute_arrow_tunnel(cls, ctx, op):
        from odps import ODPS
        from odps.tunnel import TableTunnel
        import pyarrow as pa
        import pandas as pd

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        t = o.get_table(op.table_name)
        schema_name = t.get_schema().name if t.get_schema() is not None else None
        tunnel = TableTunnel(o, project=t.project)

        if op.partition_spec is not None:
            upload_session = tunnel.create_upload_session(
                t.name, partition_spec=op.partition_spec
            )
        else:
            upload_session = tunnel.create_upload_session(t.name, schema=schema_name)

        logger.debug(
            "Start writing table %s split index: %s", op.table_name, op.inputs[0].index
        )
        writer = upload_session.open_arrow_writer(0)
        arrow_rb = pa.RecordBatch.from_pandas(ctx[op.inputs[0].key])
        writer.write(arrow_rb)
        writer.close()
        upload_session.commit([0])
        logger.debug(
            "Finish writing table %s split index: %s", op.table_name, op.inputs[0].index
        )
        ctx[op.outputs[0].key] = pd.DataFrame()

    @classmethod
    def execute(cls, ctx, op):
        if op.cupid_handle is not None:
            cls._execute_in_cupid(ctx, op)
        else:
            cls._execute_arrow_tunnel(ctx, op)


class DataFrameWriteTableCommit(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123462

    _dtypes = SeriesField("dtypes")

    _odps_params = DictField("odps_params")
    _table_name = StringField("table_name")
    _overwrite = BoolField("overwrite")
    _blocks = DictField("blocks")
    _cupid_handle = StringField("cupid_handle")
    _is_terminal = BoolField("is_terminal")

    def __init__(
        self,
        dtypes=None,
        odps_params=None,
        table_name=None,
        blocks=None,
        cupid_handle=None,
        overwrite=False,
        is_terminal=None,
        **kw
    ):
        kw.update(_output_type_kw)
        super(DataFrameWriteTableCommit, self).__init__(
            _dtypes=dtypes,
            _odps_params=odps_params,
            _table_name=table_name,
            _blocks=blocks,
            _overwrite=overwrite,
            _cupid_handle=cupid_handle,
            _is_terminal=is_terminal,
            **kw
        )

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def table_name(self):
        return self._table_name

    @property
    def blocks(self):
        return self._blocks

    @property
    def overwrite(self):
        return self._overwrite

    @property
    def cupid_handle(self):
        return self._cupid_handle

    @property
    def odps_params(self):
        return self._odps_params

    @property
    def is_terminal(self):
        return self._is_terminal

    @classmethod
    def execute(cls, ctx, op):
        import pandas as pd
        from odps import ODPS
        from odps.accounts import BearerTokenAccount
        from cupid import CupidSession, context
        from cupid.io.table import CupidTableUploadSession

        if op.is_terminal:
            bearer_token = context().get_bearer_token()
            account = BearerTokenAccount(bearer_token)
            project = os.environ.get("ODPS_PROJECT_NAME", None)
            odps_params = op.odps_params.copy()
            if project:
                odps_params["project"] = project
            endpoint = (
                os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
            )
            o = ODPS(
                None,
                None,
                account=account,
                project=odps_params["project"],
                endpoint=endpoint,
            )
            cupid_session = CupidSession(o)

            project_name, table_name = op.table_name.split(".")
            if table_name.startswith('`') and table_name.endswith('`'):
                table_name = table_name[1:-1]
            upload_session = CupidTableUploadSession(
                session=cupid_session,
                table_name=table_name,
                project_name=project_name,
                handle=op.cupid_handle,
                blocks=op.blocks,
            )
            upload_session.commit(overwrite=op.overwrite)

        ctx[op.outputs[0].key] = pd.DataFrame()


def write_odps_table(
    df,
    table,
    partition=None,
    overwrite=False,
    unknown_as_string=None,
    odps_params=None,
    write_batch_size=None,
):
    op = DataFrameWriteTable(
        dtypes=df.dtypes,
        odps_params=odps_params,
        table_name=table.full_table_name,
        unknown_as_string=unknown_as_string,
        partition_spec=partition,
        overwrite=overwrite,
        write_batch_size=write_batch_size,
    )
    return op(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import sys
import time

import requests

from .. import errors, options, serializers, types, utils
from ..compat import Enum, six
from ..lib.monotonic import monotonic
from ..models import Projects, Record, TableSchema
from ..types import Column
from .base import BaseTunnel, TUNNEL_VERSION
from .errors import TunnelError, TunnelWriteTimeout
from .io.reader import TunnelRecordReader, TunnelArrowReader, ArrowRecordReader
from .io.stream import CompressOption, get_decompress_stream
from .io.writer import (
    RecordWriter,
    BufferedArrowWriter,
    BufferedRecordWriter,
    StreamRecordWriter,
    ArrowWriter,
    Upsert,
)

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pyarrow as pa
except ImportError:
    pa = None

logger = logging.getLogger(__name__)
TUNNEL_DATA_TRANSFORM_VERSION = "v1"
DEFAULT_UPSERT_COMMIT_TIMEOUT = 120


def _wrap_upload_call(request_id):
    def wrapper(func):
        @six.wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except requests.ConnectionError as ex:
                ex_str = str(ex)
                if 'timed out' in ex_str:
                    raise TunnelWriteTimeout(ex_str, request_id=request_id)
                else:
                    raise

        return wrapped

    return wrapper


class BaseTableTunnelSession(serializers.JSONSerializableModel):
    @staticmethod
    def get_common_headers(content_length=None, chunked=False):
        header = {
            "odps-tunnel-date-transform": TUNNEL_DATA_TRANSFORM_VERSION,
            "x-odps-tunnel-version": TUNNEL_VERSION,
        }
        if content_length is not None:
            header["Content-Length"] = content_length
        if chunked:
            header.update(
                {
                    "Transfer-Encoding": "chunked",
                    "Content-Type": "application/octet-stream",
                }
            )
        return header

    @staticmethod
    def normalize_partition_spec(partition_spec):
        if isinstance(partition_spec, six.string_types):
            partition_spec = types.PartitionSpec(partition_spec)
        if isinstance(partition_spec, types.PartitionSpec):
            partition_spec = str(partition_spec).replace("'", '')
        return partition_spec

    def get_common_params(self, **kwargs):
        params = {k: str(v) for k, v in kwargs.items()}
        if getattr(self, "_quota_name", None):
            params['quotaName'] = self._quota_name
        if self._partition_spec is not None and len(self._partition_spec) > 0:
            params['partition'] = self._partition_spec
        return params

    def check_tunnel_response(self, resp):
        if not self._client.is_ok(resp):
            e = TunnelError.parse(resp)
            raise e

    def new_record(self, values=None):
        return Record(
            schema=self.schema, values=values,
            max_field_size=getattr(self, "max_field_size", None),
        )


class TableDownloadSession(BaseTableTunnelSession):
    __slots__ = (
        '_client', '_table', '_partition_spec', '_compress_option', '_quota_name'
    )

    class Status(Enum):
        Unknown = 'UNKNOWN'
        Normal = 'NORMAL'
        Closes = 'CLOSES'
        Expired = 'EXPIRED'
        Initiating = 'INITIATING'

    id = serializers.JSONNodeField('DownloadID')
    status = serializers.JSONNodeField(
        'Status', parse_callback=lambda s: TableDownloadSession.Status(s.upper())
    )
    count = serializers.JSONNodeField('RecordCount')
    schema = serializers.JSONNodeReferenceField(TableSchema, 'Schema')
    quota_name = serializers.JSONNodeField('QuotaName')

    def __init__(self, client, table, partition_spec, download_id=None,
                 compress_option=None, async_mode=True, timeout=None,
                 quota_name=None, **kw):
        super(TableDownloadSession, self).__init__()

        self._client = client
        self._table = table
        self._partition_spec = self.normalize_partition_spec(partition_spec)

        self._quota_name = quota_name

        if "async_" in kw:
            async_mode = kw.pop("async_")
        if kw:
            raise TypeError("Cannot accept arguments %s" % ", ".join(kw.keys()))
        if download_id is None:
            self._init(async_mode=async_mode, timeout=timeout)
        else:
            self.id = download_id
            self.reload()
        self._compress_option = compress_option

        logger.info("Tunnel session created: %r", self)
        if options.tunnel_session_create_callback:
            options.tunnel_session_create_callback(self)

    def __repr__(self):
        return "<TableDownloadSession id=%s project=%s table=%s partition_spec=%s>" % (
            self.id,
            self._table.project.name,
            self._table.name,
            self._partition_spec,
        )

    def _init(self, async_mode, timeout):
        params = self.get_common_params(downloads='')
        headers = self.get_common_headers(content_length=0)
        if async_mode:
            params['asyncmode'] = 'true'

        url = self._table.table_resource()
        try:
            resp = self._client.post(url, {}, params=params, headers=headers, timeout=timeout)
        except requests.exceptions.ReadTimeout:
            if callable(options.tunnel_session_create_timeout_callback):
                options.tunnel_session_create_timeout_callback(*sys.exc_info())
            raise
        self.check_tunnel_response(resp)

        delay_time = 0.1
        self.parse(resp, obj=self)
        while self.status == self.Status.Initiating:
            time.sleep(delay_time)
            delay_time = min(delay_time * 2, 5)
            self.reload()
        if self.schema is not None:
            self.schema.build_snapshot()

    def reload(self):
        params = self.get_common_params(downloadid=self.id)
        headers = self.get_common_headers(content_length=0)

        url = self._table.table_resource()
        resp = self._client.get(url, params=params, headers=headers)
        self.check_tunnel_response(resp)

        self.parse(resp, obj=self)
        if self.schema is not None:
            self.schema.build_snapshot()

    def _open_reader(
        self, start, count, compress=False, columns=None, arrow=False, reader_cls=None, **kw
    ):
        compress_option = self._compress_option or CompressOption()

        actions = ["data"]
        params = self.get_common_params(downloadid=self.id)
        headers = self.get_common_headers(content_length=0)
        if compress:
            encoding = compress_option.algorithm.get_encoding()
            if encoding:
                headers['Accept-Encoding'] = encoding

        params['rowrange'] = '(%s,%s)' % (start, count)
        if columns is not None and len(columns) > 0:
            col_name = lambda col: col.name if isinstance(col, types.Column) else col
            params['columns'] = ','.join(col_name(col) for col in columns)

        if arrow:
            actions.append("arrow")

        url = self._table.table_resource()
        resp = self._client.get(url, stream=True, actions=actions, params=params, headers=headers)
        self.check_tunnel_response(resp)

        content_encoding = resp.headers.get('Content-Encoding')
        if content_encoding is not None:
            compress_algo = CompressOption.CompressAlgorithm.from_encoding(content_encoding)
            if compress_algo != compress_option.algorithm:
                compress_option = self._compress_option = CompressOption(compress_algo, -1, 0)
            compress = True
        else:
            compress = False

        option = compress_option if compress else None
        input_stream = get_decompress_stream(resp, option)
        return reader_cls(self.schema, input_stream, columns=columns, **kw)

    def open_record_reader(self, start, count, compress=False, columns=None):
        return self._open_reader(
            start, count, compress=compress, columns=columns,
            reader_cls=TunnelRecordReader, partition_spec=self._partition_spec,
        )

    if np is not None:
        @utils.survey
        def open_pandas_reader(self, start, count, compress=False, columns=None):
            from .pdio.pdreader_c import TunnelPandasReader
            return self._open_reader(start, count, compress=compress, columns=columns,
                                     reader_cls=TunnelPandasReader)

    def open_arrow_reader(self, start, count, compress=False, columns=None):
        return self._open_reader(
            start, count, compress=compress, columns=columns,
            arrow=True, reader_cls=TunnelArrowReader
        )


class TableUploadSession(BaseTableTunnelSession):
    __slots__ = (
        '_client', '_table', '_partition_spec', '_compress_option',
        '_overwrite', '_quota_name',
    )

    class Status(Enum):
        Unknown = 'UNKNOWN'
        Normal = 'NORMAL'
        Closing = 'CLOSING'
        Closed = 'CLOSED'
        Canceled = 'CANCELED'
        Expired = 'EXPIRED'
        Critical = 'CRITICAL'

    id = serializers.JSONNodeField('UploadID')
    status = serializers.JSONNodeField(
        'Status', parse_callback=lambda s: TableUploadSession.Status(s.upper())
    )
    blocks = serializers.JSONNodesField('UploadedBlockList', 'BlockID')
    schema = serializers.JSONNodeReferenceField(TableSchema, 'Schema')
    max_field_size = serializers.JSONNodeField('MaxFieldSize')
    quota_name = serializers.JSONNodeField('QuotaName')

    def __init__(self, client, table, partition_spec, upload_id=None,
                 compress_option=None, overwrite=False, quota_name=None):
        super(TableUploadSession, self).__init__()

        self._client = client
        self._table = table
        self._partition_spec = self.normalize_partition_spec(partition_spec)

        self._quota_name = quota_name
        self._overwrite = overwrite

        if upload_id is None:
            self._init()
        else:
            self.id = upload_id
            self.reload()
        self._compress_option = compress_option

        logger.info("Tunnel session created: %r", self)
        if options.tunnel_session_create_callback:
            options.tunnel_session_create_callback(self)

    def __repr__(self):
        return "<TableUploadSession id=%s project=%s table=%s partition_spec=%s>" % (
            self.id,
            self._table.project.name,
            self._table.name,
            self._partition_spec,
        )

    def _create_or_reload_session(self, reload=False):
        headers = self.get_common_headers(content_length=0)
        params = self.get_common_params(reload=reload)
        if not reload and self._overwrite:
            params["overwrite"] = "true"

        if reload:
            params['uploadid'] = self.id
        else:
            params['uploads'] = 1

        def _call_tunnel(func, *args, **kw):
            resp = func(*args, **kw)
            self.check_tunnel_response(resp)
            return resp

        url = self._table.table_resource()
        if reload:
            resp = utils.call_with_retry(
                _call_tunnel, self._client.get, url, params=params, headers=headers
            )
        else:
            resp = utils.call_with_retry(
                _call_tunnel, self._client.post, url, {}, params=params, headers=headers
            )

        self.parse(resp, obj=self)
        if self.schema is not None:
            self.schema.build_snapshot()

    def _init(self):
        self._create_or_reload_session(reload=False)

    def reload(self):
        self._create_or_reload_session(reload=True)

    def _open_writer(
        self,
        block_id=None,
        compress=False,
        buffer_size=None,
        writer_cls=None,
        initial_block_id=None,
        block_id_gen=None,
    ):
        compress_option = self._compress_option or CompressOption()

        params = self.get_common_params(uploadid=self.id)
        headers = self.get_common_headers(chunked=True)
        if compress:
            # special: rewrite LZ4 to ARROW_LZ4 for arrow tunnels
            if (
                writer_cls is not None
                and issubclass(writer_cls, ArrowWriter)
                and compress_option.algorithm == CompressOption.CompressAlgorithm.ODPS_LZ4
            ):
                compress_option.algorithm = CompressOption.CompressAlgorithm.ODPS_ARROW_LZ4
            encoding = compress_option.algorithm.get_encoding()
            if encoding:
                headers['Content-Encoding'] = encoding

        url = self._table.table_resource()
        option = compress_option if compress else None

        if block_id is None:
            @_wrap_upload_call(self.id)
            def upload_block(blockid, data):
                params['blockid'] = blockid
                return utils.call_with_retry(
                    self._client.put, url, data=data, params=params, headers=headers
                )

            if writer_cls is ArrowWriter:
                writer_cls = BufferedArrowWriter
                params['arrow'] = ''
            else:
                writer_cls = BufferedRecordWriter

            writer = writer_cls(
                self.schema,
                upload_block,
                compress_option=option,
                buffer_size=buffer_size,
                block_id=initial_block_id,
                block_id_gen=block_id_gen,
            )
        else:
            params['blockid'] = block_id

            @_wrap_upload_call(self.id)
            def upload(data):
                return self._client.put(url, data=data, params=params, headers=headers)

            if writer_cls is ArrowWriter:
                params['arrow'] = ''

            writer = writer_cls(self.schema, upload, compress_option=option)
        return writer

    def open_record_writer(
        self, block_id=None, compress=False, buffer_size=None, initial_block_id=None,
        block_id_gen=None
    ):
        return self._open_writer(
            block_id=block_id, compress=compress, buffer_size=buffer_size,
            initial_block_id=initial_block_id, block_id_gen=block_id_gen,
            writer_cls=RecordWriter
        )

    def open_arrow_writer(
        self, block_id=None, compress=False, initial_block_id=None, block_id_gen=None
    ):
        return self._open_writer(
            block_id=block_id, compress=compress, initial_block_id=initial_block_id,
            block_id_gen=block_id_gen, writer_cls=ArrowWriter
        )

    if np is not None:
        @utils.survey
        def open_pandas_writer(self, block_id=None, compress=False, buffer_size=None):
            from .pdio import TunnelPandasWriter
            return self._open_writer(block_id=block_id, compress=compress, buffer_size=buffer_size,
                                     writer_cls=TunnelPandasWriter)

    def get_block_list(self):
        self.reload()
        return self.blocks

    def commit(self, blocks):
        if blocks is None:
            raise ValueError('Invalid parameter: blocks.')
        if isinstance(blocks, six.integer_types):
            blocks = [blocks, ]

        server_block_map = dict(
            [
                (int(block_id), True) for block_id in self.get_block_list()
            ]
        )
        client_block_map = dict([(int(block_id), True) for block_id in blocks])

        if len(server_block_map) != len(client_block_map):
            raise TunnelError(
                'Blocks not match, server: %s, tunnelServerClient: %s. '
                'Make sure all block writers closed or with-blocks exited.' % (
                    len(server_block_map), len(client_block_map)
                )
            )

        for block_id in blocks:
            if block_id not in server_block_map:
                raise TunnelError(
                    'Block not exists on server, block id is %s' % (block_id,)
                )

        self._complete_upload()

    def _complete_upload(self):
        headers = self.get_common_headers()
        params = self.get_common_params(uploadid=self.id)
        url = self._table.table_resource()

        resp = utils.call_with_retry(
            self._client.post,
            url,
            '',
            params=params,
            headers=headers,
            exc_type=(requests.Timeout, requests.ConnectionError, errors.InternalServerError),
        )
        self.parse(resp, obj=self)


class Slot(object):
    def __init__(self, slot, server):
        self._slot = slot
        self._ip = None
        self._port = None
        self.set_server(server, True)

    @property
    def slot(self):
        return self._slot

    @property
    def ip(self):
        return self._ip

    @property
    def port(self):
        return self._port

    @property
    def server(self):
        return str(self._ip) + ':' + str(self._port)

    def set_server(self, server, check_empty=False):
        if len(server.split(':')) != 2:
            raise TunnelError('Invalid slot format: {}'.format(server))

        ip, port = server.split(':')

        if check_empty:
            if (not ip) or (not port):
                raise TunnelError('Empty server ip or port')
        if ip:
            self._ip = ip
        if port:
            self._port = int(port)


class TableStreamUploadSession(BaseTableTunnelSession):
    __slots__ = (
        '_client', '_table', '_partition_spec', '_compress_option', '_quota_name'
    )

    class Slots(object):
        def __init__(self, slot_elements):
            self._slots = []
            self._cur_index = -1
            for value in slot_elements:
                if len(value) != 2:
                    raise TunnelError('Invalid slot routes')
                self._slots.append(Slot(value[0], value[1]))

            if len(self._slots) > 0:
                self._cur_index = random.randint(0, len(self._slots))
            self._iter = iter(self)

        def __len__(self):
            return len(self._slots)

        def __next__(self):
            return next(self._iter)

        def __iter__(self):
            while True:
                if self._cur_index < 0:
                    yield None
                else:
                    self._cur_index += 1
                    if self._cur_index >= len(self._slots):
                        self._cur_index = 0
                    yield self._slots[self._cur_index]

    schema = serializers.JSONNodeReferenceField(TableSchema, 'schema')
    id = serializers.JSONNodeField('session_name')
    status = serializers.JSONNodeField('status')
    slots = serializers.JSONNodeField(
        'slots', parse_callback=lambda val: TableStreamUploadSession.Slots(val))
    quota_name = serializers.JSONNodeField('QuotaName')

    def __init__(
        self, client, table, partition_spec, compress_option=None, quota_name=None
    ):
        super(TableStreamUploadSession, self).__init__()

        self._client = client
        self._table = table
        self._partition_spec = self.normalize_partition_spec(partition_spec)

        self._quota_name = quota_name

        self._init()
        self._compress_option = compress_option

        logger.info("Tunnel session created: %r", self)
        if options.tunnel_session_create_callback:
            options.tunnel_session_create_callback(self)

    def __repr__(self):
        return (
            "<TableStreamUploadSession id=%s project=%s table=%s partition_spec=%s>"
            % (
                self.id,
                self._table.project.name,
                self._table.name,
                self._partition_spec,
            )
        )

    def _init(self):
        params = self.get_common_params()
        headers = self.get_common_headers(content_length=0)

        url = self._get_resource()
        resp = self._client.post(url, {}, params=params, headers=headers)
        self.check_tunnel_response(resp)

        self.parse(resp, obj=self)
        if self.schema is not None:
            self.schema.build_snapshot()

    def _get_resource(self):
        return self._table.table_resource() + '/streams'

    def reload(self):
        params = self.get_common_params(uploadid=self.id)
        headers = self.get_common_headers(content_length=0)

        url = self._get_resource()
        resp = self._client.get(url, params=params, headers=headers)
        self.check_tunnel_response(resp)

        self.parse(resp, obj=self)
        if self.schema is not None:
            self.schema.build_snapshot()

    def abort(self):
        params = self.get_common_params(uploadid=self.id)

        slot = next(iter(self.slots))
        headers = self.get_common_headers(content_length=0)
        headers["odps-tunnel-routed-server"] = slot.server

        url = self._get_resource()
        resp = self._client.post(url, {}, params=params, headers=headers)
        self.check_tunnel_response(resp)

    def reload_slots(self, slot, server, slot_num):
        if len(self.slots) != slot_num:
            self.reload()
        else:
            slot.set_server(server)

    def _open_writer(self, compress=False):
        compress_option = self._compress_option or CompressOption()

        slot = next(iter(self.slots))

        headers = self.get_common_headers(chunked=True)
        headers.update({
            "odps-tunnel-slot-num": str(len(self.slots)),
            "odps-tunnel-routed-server": slot.server,
        })

        if compress:
            encoding = compress_option.algorithm.get_encoding()
            if encoding:
                headers['Content-Encoding'] = encoding

        params = self.get_common_params(uploadid=self.id, slotid=slot.slot)
        url = self._get_resource()
        option = compress_option if compress else None

        @_wrap_upload_call(self.id)
        def upload_block(data):
            return self._client.put(url, data=data, params=params, headers=headers)

        writer = StreamRecordWriter(
            self.schema, upload_block, session=self, slot=slot, compress_option=option
        )

        return writer

    def open_record_writer(self, compress=False):
        return self._open_writer(compress=compress)


class TableUpsertSession(BaseTableTunnelSession):
    __slots__ = (
        '_client', '_table', '_partition_spec', '_compress_option',
        '_slot_num', '_commit_timeout', '_quota_name'
    )

    UPSERT_EXTRA_COL_NUM = 5
    UPSERT_VERSION_KEY = "__version"
    UPSERT_APP_VERSION_KEY = "__app_version"
    UPSERT_OPERATION_KEY = "__operation"
    UPSERT_KEY_COLS_KEY = "__key_cols"
    UPSERT_VALUE_COLS_KEY = "__value_cols"

    class Status(Enum):
        Normal = "NORMAL"
        Committing = "COMMITTING"
        Committed = "COMMITTED"
        Expired = "EXPIRED"
        Critical = "CRITICAL"
        Aborted = "ABORTED"

    class Slots(object):
        def __init__(self, slot_elements):
            self._slots = []
            self._buckets = {}
            for value in slot_elements:
                slot = Slot(value['slot_id'], value['worker_addr'])
                self._slots.append(slot)
                self._buckets.update({idx: slot for idx in value['buckets']})

            for idx in self._buckets.keys():
                if idx > len(self._buckets):
                    raise TunnelError("Invalid bucket value: " + str(idx))

        @property
        def buckets(self):
            return self._buckets

        def __len__(self):
            return len(self._slots)

    schema = serializers.JSONNodeReferenceField(TableSchema, 'schema')
    id = serializers.JSONNodeField('id')
    status = serializers.JSONNodeField(
        'status', parse_callback=lambda s: TableUpsertSession.Status(s.upper())
    )
    slots = serializers.JSONNodeField(
        'slots', parse_callback=lambda val: TableUpsertSession.Slots(val))
    quota_name = serializers.JSONNodeField('quota_name')
    hash_keys = serializers.JSONNodeField('hash_key')
    hasher = serializers.JSONNodeField('hasher')

    def __init__(
        self, client, table, partition_spec, compress_option=None, slot_num=1,
        commit_timeout=DEFAULT_UPSERT_COMMIT_TIMEOUT, quota_name=None
    ):
        super(TableUpsertSession, self).__init__()

        self._client = client
        self._table = table
        self._partition_spec = self.normalize_partition_spec(partition_spec)
        self._quota_name = quota_name

        self._slot_num = slot_num
        self._commit_timeout = commit_timeout

        self._init()
        self._compress_option = compress_option

        logger.info("Upsert session created: %r", self)
        if options.tunnel_session_create_callback:
            options.tunnel_session_create_callback(self)

    def __repr__(self):
        return "<TableUpsertSession id=%s project=%s table=%s partition_spec=%s>" % (
            self.id,
            self._table.project.name,
            self._table.name,
            self._partition_spec,
        )

    @property
    def endpoint(self):
        return self._client.endpoint

    @property
    def buckets(self):
        return self.slots.buckets

    def _get_resource(self):
        return self._table.table_resource() + '/upserts'

    def _patch_schema(self):
        if self.schema is None:
            return
        patch_schema = types.OdpsSchema(
            [
                Column(self.UPSERT_VERSION_KEY, "bigint"),
                Column(self.UPSERT_APP_VERSION_KEY, "bigint"),
                Column(self.UPSERT_OPERATION_KEY, "tinyint"),
                Column(self.UPSERT_KEY_COLS_KEY, "array<bigint>"),
                Column(self.UPSERT_VALUE_COLS_KEY, "array<bigint>"),
            ],
        )
        self.schema = self.schema.extend(patch_schema)
        self.schema.build_snapshot()

    def _init_or_reload(self, reload=False):
        params = self.get_common_params()
        headers = self.get_common_headers(content_length=0)

        if not reload:
            params['slotnum'] = str(self._slot_num)
        else:
            params['upsertid'] = self.id

        url = self._get_resource()
        if not reload:
            resp = self._client.post(url, {}, params=params, headers=headers)
        else:
            resp = self._client.get(url, params=params, headers=headers)
        if self._client.is_ok(resp):
            self.parse(resp, obj=self)
            self._patch_schema()
        else:
            e = TunnelError.parse(resp)
            raise e

    def _init(self):
        self._init_or_reload()

    def new_record(self, values=None):
        if values:
            values = list(values) + [None] * 5
        return super(TableUpsertSession, self).new_record(values)

    def reload(self, init=False):
        self._init_or_reload(reload=True)

    def abort(self):
        params = self.get_common_params(upsertid=self.id)
        headers = self.get_common_headers(content_length=0)
        headers["odps-tunnel-routed-server"] = self.slots.buckets[0].server

        url = self._get_resource()
        resp = self._client.delete(url, params=params, headers=headers)
        self.check_tunnel_response(resp)

    def open_upsert_stream(self, compress=False):
        params = self.get_common_params(upsertid=self.id)
        headers = self.get_common_headers()

        compress_option = self._compress_option or CompressOption()
        if compress:
            encoding = compress_option.algorithm.get_encoding()
            if encoding:
                headers['Content-Encoding'] = encoding

        url = self._get_resource()

        @_wrap_upload_call(self.id)
        def upload_block(bucket, slot, record_count, data):
            req_params = params.copy()
            req_params.update(
                dict(bucketid=bucket, slotid=str(slot.slot), record_count=str(record_count))
            )
            req_headers = headers.copy()
            req_headers["odps-tunnel-routed-server"] = slot.server
            req_headers["Content-Length"] = len(data)
            return self._client.put(url, data=data, params=req_params, headers=req_headers)

        return Upsert(self.schema, upload_block, self, compress_option)

    def commit(self, async_=False):
        params = self.get_common_params(upsertid=self.id)
        headers = self.get_common_headers(content_length=0)
        headers["odps-tunnel-routed-server"] = self.slots.buckets[0].server

        url = self._get_resource()
        resp = self._client.post(url, params=params, headers=headers)
        self.check_tunnel_response(resp)
        self.reload()

        if async_:
            return

        delay = 1
        start = monotonic()
        while self.status in (TableUpsertSession.Status.Committing, TableUpsertSession.Status.Normal):
            try:
                if monotonic() - start > self._commit_timeout:
                    raise TunnelError("Commit session timeout")
                time.sleep(delay)

                resp = self._client.post(url, params=params, headers=headers)
                self.check_tunnel_response(resp)
                self.reload()

                delay = min(8, delay * 2)
            except (errors.StreamSessionNotFound, errors.UpsertSessionNotFound):
                self.status = TableUpsertSession.Status.Committed
        if self.status != TableUpsertSession.Status.Committed:
            raise TunnelError("commit session failed, status: " + self.status.value)


class TableTunnel(BaseTunnel):
    def _get_tunnel_table(self, table, schema=None):
        if not isinstance(table, six.string_types):
            schema = schema or getattr(table.get_schema(), "name", None)
            table = table.name
        parent = Projects(client=self.tunnel_rest)[self._project.name]
        # tailor project for resource locating only
        parent._set_tunnel_defaults()
        if schema is not None:
            parent = parent.schemas[schema]
        return parent.tables[table]

    @staticmethod
    def _build_compress_option(compress_algo=None, level=None, strategy=None):
        if compress_algo is None:
            return None
        return CompressOption(
            compress_algo=compress_algo, level=level, strategy=strategy
        )

    def create_download_session(self, table, async_mode=False, partition_spec=None,
                                download_id=None, compress_option=None,
                                compress_algo=None, compress_level=None,
                                compress_strategy=None, schema=None, timeout=None, **kw):
        table = self._get_tunnel_table(table, schema)
        compress_option = compress_option or self._build_compress_option(
            compress_algo=compress_algo, level=compress_level, strategy=compress_strategy
        )
        if "async_" in kw:
            async_mode = kw.pop("async_")
        if kw:
            raise TypeError("Cannot accept arguments %s" % ", ".join(kw.keys()))
        return TableDownloadSession(
            self.tunnel_rest,
            table,
            partition_spec,
            download_id=download_id,
            compress_option=compress_option,
            async_mode=async_mode,
            timeout=timeout,
            quota_name=self._quota_name,
        )

    def create_upload_session(
        self,
        table,
        partition_spec=None,
        upload_id=None,
        compress_option=None,
        compress_algo=None,
        compress_level=None,
        compress_strategy=None,
        schema=None,
        overwrite=False,
    ):
        table = self._get_tunnel_table(table, schema)
        compress_option = compress_option
        compress_option = compress_option or self._build_compress_option(
            compress_algo=compress_algo, level=compress_level, strategy=compress_strategy
        )
        return TableUploadSession(
            self.tunnel_rest,
            table,
            partition_spec,
            upload_id=upload_id,
            compress_option=compress_option,
            overwrite=overwrite,
            quota_name=self._quota_name,
        )

    def create_stream_upload_session(
        self,
        table,
        partition_spec=None,
        compress_option=None,
        compress_algo=None,
        compress_level=None,
        compress_strategy=None,
        schema=None,
    ):
        table = self._get_tunnel_table(table, schema)
        compress_option = compress_option or self._build_compress_option(
            compress_algo=compress_algo, level=compress_level, strategy=compress_strategy
        )
        return TableStreamUploadSession(
            self.tunnel_rest,
            table,
            partition_spec,
            compress_option=compress_option,
            quota_name=self._quota_name,
        )

    def create_upsert_session(
        self,
        table,
        partition_spec=None,
        slot_num=1,
        commit_timeout=120,
        compress_option=None,
        compress_algo=None,
        compress_level=None,
        compress_strategy=None,
        schema=None,
    ):
        table = self._get_tunnel_table(table, schema)
        compress_option = compress_option or self._build_compress_option(
            compress_algo=compress_algo, level=compress_level, strategy=compress_strategy
        )
        return TableUpsertSession(
            self.tunnel_rest,
            table,
            partition_spec,
            slot_num=slot_num,
            commit_timeout=commit_timeout,
            compress_option=compress_option,
            quota_name=self._quota_name,
        )

    def open_preview_reader(
        self,
        table,
        partition_spec=None,
        columns=None,
        limit=None,
        compress_option=None,
        compress_algo=None,
        compress_level=None,
        compress_strategy=None,
        arrow=True,
        timeout=None,
        make_compat=True,
        read_all=False,
    ):
        if pa is None:
            raise ImportError("Need pyarrow to run open_preview_reader.")

        tunnel_table = self._get_tunnel_table(table)
        compress_option = compress_option or self._build_compress_option(
            compress_algo=compress_algo, level=compress_level, strategy=compress_strategy
        )

        params = {"limit": str(limit) if limit else "-1"}
        partition_spec = BaseTableTunnelSession.normalize_partition_spec(partition_spec)
        if columns:
            col_set = set(columns)
            ordered_col = [c.name for c in table.table_schema if c.name in col_set]
            params["columns"] = ",".join(ordered_col)
        if partition_spec is not None and len(partition_spec) > 0:
            params['partition'] = partition_spec

        headers = BaseTableTunnelSession.get_common_headers(content_length=0)
        if compress_option:
            encoding = compress_option.algorithm.get_encoding(legacy=False)
            if encoding:
                headers['Accept-Encoding'] = encoding

        url = tunnel_table.table_resource(force_schema=True) + "/preview"
        resp = self.tunnel_rest.get(
            url, stream=True, params=params, headers=headers, timeout=timeout
        )
        if not self.tunnel_rest.is_ok(resp):  # pragma: no cover
            e = TunnelError.parse(resp)
            raise e

        input_stream = get_decompress_stream(resp)
        if input_stream.peek() is None:
            # stream is empty, replace with empty stream
            input_stream = None

        reader = TunnelArrowReader(
            table.table_schema, input_stream, columns=columns, use_ipc_stream=True
        )
        if not arrow:
            reader = ArrowRecordReader(
                reader, make_compat=make_compat, read_all=read_all
            )
        return reader



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import itertools
import weakref
from collections import deque, defaultdict, OrderedDict

from ...compat import six, Iterable
from ...dag import DAG, Queue
from . import utils


class NodeMetaclass(type):
    def __new__(mcs, name, bases, kv):
        mixin_slots = []
        for b in bases:
            mixin_slots.extend(list(getattr(b, '_mixin_slots', [])))

        if kv.pop('_add_args_slots', True):
            slots = mixin_slots + list(kv.get('__slots__', [])) + list(kv.get('_mixin_slots', [])) + \
                    list(kv.get('_slots', []))
            args = kv.get('_args', [])
            slots.extend(args)
            slots = OrderedDict.fromkeys(slots)

            kv['__slots__'] = tuple(slot for slot in slots if slot not in kv)

        if '__slots__' not in kv:
            kv['__slots__'] = ()
        if mixin_slots:
            kv['_mixin_slots'] = []
        return type.__new__(mcs, name, bases, kv)

    def __instancecheck__(self, instance):
        from .expressions import CollectionExpr

        if issubclass(type(instance), CollectionExpr) and \
                hasattr(instance, '_proxy') and instance._proxy is not None:
            return super(NodeMetaclass, self).__instancecheck__(instance._proxy)

        return super(NodeMetaclass, self).__instancecheck__(instance)


class Node(six.with_metaclass(NodeMetaclass)):
    __slots__ = '_args_indexes', '__weakref__'
    _args = ()  # the child(ren) of the current node
    _extra_args = ()
    _non_table = False

    def __init__(self, *args, **kwargs):
        self._args_indexes = dict((name, i) for i, name in enumerate(self._args))
        self._init(*args, **kwargs)

    def _init(self, *args, **kwargs):
        for arg, value in zip(self._args, args):
            setattr(self, arg, value)

        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    @property
    def _node_id(self):
        return id(self)

    def _init_attr(self, attr, val):
        if not hasattr(self, attr):
            setattr(self, attr, val)

    @property
    def args(self):
        return tuple(getattr(self, arg, None) for arg in self._args)

    def iter_args(self):
        for name, arg in zip(self._args, self.args):
            yield name, arg

    @property
    def extra_args(self):
        return tuple(getattr(self, extra_arg, None)
                     for extra_arg in self._extra_args)

    def arg_name_values(self, extra=False):
        arg_names = self._args if not extra else self._args + self._extra_args
        arg_values = self.args if not extra else self.args + self.extra_args
        for arg_name, arg in zip(arg_names, arg_values):
            yield arg_name, arg

    def _data_source(self):
        yield None

    def data_source(self):
        stop_cond = lambda x: hasattr(x, '_source_data') and x._source_data is not None
        for n in self.traverse(top_down=True, unique=True, stop_cond=stop_cond):
            for ds in n._data_source():
                if ds is not None:
                    yield ds

    def substitute(self, old_arg, new_arg, dag=None):
        if dag is not None:
            dag.substitute(old_arg, new_arg, parents=[self])
            return

        if hasattr(old_arg, '_name') and old_arg._name is not None and \
                        new_arg._name is None:
            new_arg._name = old_arg._name

        for arg_name, arg in self.arg_name_values(extra=True):
            if not isinstance(arg, (list, tuple)):
                if arg is old_arg:
                    setattr(self, arg_name, new_arg)
            else:
                subs = list(arg)
                for i in range(len(subs)):
                    if subs[i] is old_arg:
                        subs[i] = new_arg
                setattr(self, arg_name, type(arg)(subs))

    def children(self, extra=False):
        args = []

        all_args = self.args if not extra else self.args + self.extra_args
        for arg in all_args:
            if isinstance(arg, (list, tuple)):
                args.extend(arg)
            else:
                args.append(arg)

        return [arg for arg in args if arg is not None]

    def leaves(self):
        for n in self.traverse(unique=True):
            if len(n.children()) == 0:
                yield n

    def traverse(self, top_down=False, unique=False, traversed=None,
                 extra=False, stop_cond=None):
        traversed = traversed if traversed is not None else set()

        def is_trav(n):
            if not unique:
                return False
            if n._node_id in traversed:
                return True
            traversed.add(n._node_id)
            return False

        q = deque()
        q.append(self)
        if is_trav(self):
            return

        checked = set()
        yields = set()
        while len(q) > 0:
            curr = q.popleft()
            if top_down:
                yield curr
                if stop_cond is None or not stop_cond(curr):
                    children = [c for c in curr.children(extra=extra)
                                if not is_trav(c)]
                    q.extendleft(children[::-1])
            else:
                if curr._node_id not in checked:
                    children = curr.children(extra=extra)
                    if len(children) == 0:
                        yield curr
                    else:
                        q.appendleft(curr)
                        if stop_cond is None or not stop_cond(curr):
                            q.extendleft([c for c in children
                                          if not is_trav(c) or c._node_id not in checked][::-1])
                    checked.add(curr._node_id)
                else:
                    if curr._node_id not in yields:
                        yield curr
                        if unique:
                            yields.add(curr._node_id)

    def __eq__(self, other):
        return self.equals(other)

    def equals(self, other):
        if other is None:
            return False

        if not isinstance(other, type(self)):
            return False

        def slot_values(obj):
            return [getattr(obj, slot, None) for slot in utils.get_attrs(obj)]

        def cmp(x, y):
            if isinstance(x, Node):
                res = x.equals(y)
            elif isinstance(y, Node):
                res = y.equals(y)
            elif isinstance(x, (tuple, list)) and \
                    isinstance(y, (tuple, list)):
                res = all(map(cmp, x, y))
            else:
                res = x == y
            return res
        return all(map(cmp, slot_values(self), slot_values(other)))

    def __hash__(self):
        return hash((type(self), tuple(self.children())))

    def is_ancestor(self, other):
        other = utils.get_proxied_expr(other)
        for n in self.traverse(top_down=True, unique=True):
            if n is other:
                return True
        return False

    def path(self, other, strict=False):
        all_apaths = self.all_path(other, strict=strict)
        try:
            return next(all_apaths)
        except StopIteration:
            return

    def _all_path(self, other):
        if self is other:
            yield [self, ]

        node_poses = defaultdict(lambda: 0)

        q = deque()
        q.append(self)

        while len(q) > 0:
            curr = q[-1]
            children = curr.children()

            pos = node_poses[curr._node_id]
            if len(children) == 0 or pos >= len(children):
                q.pop()
                # TODO: add this in the future
                # if pos >= len(children):
                #     node_poses[curr._node_id] = 0
                continue

            n = children[pos]
            q.append(n)
            if n is other:
                yield list(q)
                q.pop()

            node_poses[curr._node_id] += 1

    def all_path(self, other, strict=False):
        # remember, if the node has been changed into another one during traversing
        # the modification may not be applied to the paths

        i = 0
        for i, path in zip(itertools.count(1), self._all_path(other)):
            yield path

        if i == 0 and not strict:
            for path in other._all_path(self):
                yield path

    def _attr_dict(self):
        slots = utils.get_attrs(self)

        return dict((attr, getattr(self, attr, None)) for attr in slots)

    def _copy_type(self):
        return type(self)

    def copy(self, clear_keys=None, **kw):
        proxied = utils.get_proxied_expr(self)
        attr_dict = proxied._attr_dict()
        if clear_keys is not None:
            for k in clear_keys:
                attr_dict.pop(k, None)
        attr_dict.update(kw)
        copied = type(proxied)(**attr_dict)
        return copied

    def copy_to(self, target):
        slots = utils.get_attrs(self)
        for attr in slots:
            if hasattr(self, attr):
                setattr(target, attr, getattr(self, attr, None))

    def copy_tree(self, on_copy=None, extra=False, stop_cond=None):
        if on_copy is not None and not isinstance(on_copy, Iterable):
            on_copy = [on_copy, ]

        expr_id_to_copied = dict()

        def get(n):
            if n is None:
                return n
            try:
                return expr_id_to_copied[n._node_id]
            except KeyError:
                if stop_cond is not None:
                    return n
                raise

        for node in self.traverse(unique=True, extra=extra, stop_cond=stop_cond):
            node = utils.get_proxied_expr(node)
            attr_dict = node._attr_dict()

            for arg_name, arg in node.arg_name_values(extra=extra):
                if isinstance(arg, (tuple, list)):
                    attr_dict[arg_name] = type(arg)(get(it) for it in arg)
                else:
                    attr_dict[arg_name] = get(arg)

            copied_node = type(node)(**attr_dict)
            expr_id_to_copied[node._node_id] = copied_node
            if on_copy is not None:
                [func(node, copied_node) for func in on_copy]

        return expr_id_to_copied[self._node_id]

    def to_dag(self, copy=True, on_copy=None, dag=None, validate=True, stop_cond=None):
        if copy:
            expr = self.copy_tree(on_copy=on_copy, extra=True)
        else:
            expr = self

        dag = dag or ExprDAG(expr)
        queue = Queue()

        if not dag.contains_node(expr):
            dag.add_node(expr)
            queue.put(expr)

        traversed = set()
        traversed.add(expr._node_id)

        while not queue.empty():
            node = queue.get()

            for child in node.children(extra=True):
                if not dag.contains_node(child):
                    dag.add_node(child)
                if not dag.contains_edge(child, node):
                    dag.add_edge(child, node, validate=False)

                if child._node_id in traversed:
                    continue
                traversed.add(child._node_id)
                if stop_cond is None or not stop_cond(child):
                    queue.put(child)

        if validate:
            dag._validate()  # validate the DAG
        return dag

    def __getstate__(self):
        slots = utils.get_attrs(self)

        return tuple((slot, object.__getattribute__(self, slot)) for slot in slots
                     if not slot.startswith('__'))

    def __setstate__(self, state):
        self.__init__(**dict(state))


def _extract_df_inputs(o):
    if isinstance(o, Node):
        yield o
    elif isinstance(o, dict):
        for v in itertools.chain(*(_extract_df_inputs(dv) for dv in six.itervalues(o))):
            if v is not None:
                yield v
    elif isinstance(o, (list, set, tuple)):
        for v in itertools.chain(*(_extract_df_inputs(dv) for dv in o)):
            if v is not None:
                yield v
    else:
        yield None


class ExprProxy(object):
    def __init__(self, expr, d=None, compare=False):
        if d is not None:
            def callback(_):
                if self in d:
                    del d[self]
        else:
            callback = None
        self._ref = weakref.ref(expr, callback)
        self._cmp = compare
        self._hash = hash(expr)
        self._expr_id = expr._node_id

    def __call__(self):
        return self._ref()

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if isinstance(other, ExprProxy):
            if self._ref() is not None and other() is not None:
                return self._ref() is other()
            return self._expr_id == other._expr_id

        obj = self._ref()
        if obj is not None and self._cmp:
            return obj.equals(other)

        return self._expr_id == other._node_id


class ExprDAG(DAG):
    def __init__(self, root, dag=None):
        self._root = weakref.ref(root)
        super(ExprDAG, self).__init__()
        if dag is not None:
            self._graph = dag._graph
            self._map = dag._map

    @property
    def root(self):
        return self._root()

    @root.setter
    def root(self, root):
        self._root = weakref.ref(root)

    def ensure_all_nodes_in_dag(self):
        self.root.to_dag(copy=False, dag=self, validate=False)

    def traverse(self, root=None, top_down=False, traversed=None, stop_cond=None):
        root = root or self.root
        return root.traverse(top_down=top_down, unique=True,
                             traversed=traversed, stop_cond=stop_cond)

    def substitute(self, expr, new_expr, parents=None):
        if expr is self.root:
            self.root = new_expr

        parents = self.successors(expr) if parents is None else parents

        if expr._need_cache:
            new_expr.cache()

        q = Queue()
        q.put(new_expr)

        while not q.empty():
            node = q.get()
            if not self.contains_node(node):
                self.add_node(node)

            for child in node.children():
                if not self.contains_node(child):
                    q.put(child)
                    self.add_node(child)
                if not self.contains_edge(child, node):
                    self.add_edge(child, node, validate=False)

        for parent in parents:
            parent.substitute(expr, new_expr)
            self.add_edge(new_expr, parent, validate=False)
            try:
                self.remove_edge(expr, parent)
            except KeyError:
                pass

    def prune(self):
        while True:
            nodes = [n for n, succ in six.iteritems(self._graph) if len(succ) == 0]
            if len(nodes) == 1 and nodes[0] is self.root:
                break
            for node in nodes:
                if node is not self.root:
                    self.remove_node(node)

    def closest_ancestors(self, node, cond):
        collected = set()
        stop_cond = lambda n: not collected.intersection(self.successors(n))
        for n in self.bfs(node, self.predecessors, stop_cond):
            if cond(n):
                collected.add(n)
                yield n


class ExprDictionary(dict):
    def _ref(self, obj, ref_self=False):
        r = self if ref_self else None
        return obj if isinstance(obj, ExprProxy) else ExprProxy(obj, d=r)

    def __getitem__(self, item):
        if item is None:
            raise KeyError
        return dict.__getitem__(self, self._ref(item))

    def __setitem__(self, key, value):
        if key is None:
            raise KeyError
        return dict.__setitem__(self, self._ref(key, True), value)

    def __iter__(self):
        for k in dict.__iter__(self):
            yield k()

    def __delitem__(self, key):
        if key is None:
            raise KeyError
        return dict.__delitem__(self, self._ref(key))

    def __contains__(self, item):
        if item is None:
            return False
        return dict.__contains__(self, self._ref(item))

    def get(self, k, d=None):
        if k is None:
            return d
        return dict.get(self, self._ref(k), d)

    def has_key(self, k):
        if k is None:
            return False
        return dict.has_key(self, self._ref(k))

    def pop(self, k, d=None):
        if k is None:
            return False
        return dict.pop(self, self._ref(k), d)

    def popitem(self):
        k, v = dict.popitem(self)
        return k(), v

    def setdefault(self, k, d=None):
        if k is None:
            raise KeyError
        return dict.setdefault(self, self._ref(k, True), d)

    def update(self, E=None, **F):
        if hasattr(E, 'keys'):
            for k in E.keys():
                self[k] = E[k]
        elif E is not None:
            for k, v in E:
                self[k] = v
        else:
            for k, v in six.iteritems(F):
                self[k] = v

    if six.PY2:
        def items(self):
            return [(k(), v) for k, v in dict.items(self)]

        def keys(self):
            return [k() for k in dict.keys(self)]

        def iteritems(self):
            for k, v in dict.iteritems(self):
                yield k(), v

        def iterkeys(self):
            for k in dict.iterkeys(self):
                yield k()
    else:
        def items(self):
            for k, v in dict.items(self):
                yield k(), v

        def keys(self):
            for k in dict.keys(self):
                yield k()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging
import tempfile
import subprocess

import numpy as np
from mars.serialize import BytesField, ListField, Int32Field, StringField
from mars.utils import to_binary


logger = logging.getLogger(__name__)

try:
    from mars.remote.run_script import RunScript
except ImportError:
    try:
        from mars.learn.operands import LearnMergeDictOperand, OutputType
    except ImportError:
        from mars.operands import MergeDictOperand as LearnMergeDictOperand
        from mars.operands import OutputType

    class RunScript(LearnMergeDictOperand):
        _op_type_ = 743210

        _code = BytesField("code")
        _mode = StringField("mode")
        _command_args = ListField("command_args")
        _world_size = Int32Field("world_size")
        _rank = Int32Field("rank")

        def __init__(
            self,
            code=None,
            mode=None,
            command_args=None,
            world_size=None,
            rank=None,
            merge=None,
            output_types=None,
            **kw
        ):
            super().__init__(
                _code=code,
                _mode=mode,
                _command_args=command_args,
                _world_size=world_size,
                _rank=rank,
                _merge=merge,
                _output_types=output_types,
                **kw
            )
            if self._output_types is None:
                self._output_types = [OutputType.object]

        @property
        def code(self):
            return self._code

        @property
        def mode(self):
            return self._mode

        @property
        def world_size(self):
            return self._world_size

        @property
        def rank(self):
            return self._rank

        @property
        def command_args(self):
            return self._command_args or []

        def __call__(self):
            return self.new_tileable(None)

        @classmethod
        def tile(cls, op):
            out_chunks = []
            for i in range(op.world_size):
                chunk_op = op.copy().reset_key()
                chunk_op._rank = i
                out_chunks.append(chunk_op.new_chunk(None, index=(i,)))

            new_op = op.copy()
            return new_op.new_tileables(
                op.inputs,
                chunks=out_chunks,
                nsplits=(tuple(np.nan for _ in range(len(out_chunks))),),
            )

        @classmethod
        def _execute_with_subprocess(cls, op, env=None):
            # write source code into a temp file
            fd, filename = tempfile.mkstemp(".py")
            with os.fdopen(fd, "wb") as f:
                f.write(op.code)
            logger.debug("Write code to temp file.")

            env = env or dict()
            envs = os.environ.copy().update(env)
            try:
                # exec code in a new process
                process = subprocess.Popen(
                    [sys.executable, filename] + op.command_args, env=envs
                )
                process.wait()
                if process.returncode != 0:
                    raise RuntimeError("Run script failed")

            finally:
                os.remove(filename)

        @classmethod
        def _execute_with_exec(cls, op, local=None):
            local = local or dict()

            try:
                exec(op.code, local)
            finally:
                sys.stdout.flush()

        @classmethod
        def _set_envs(cls, ctx, op):
            scheduler_address = ctx._scheduler_address
            session_id = ctx._session_id

            # set mars envs
            env = os.environ
            env["MARS_SCHEDULER_ADDRESS"] = str(scheduler_address)
            env["MARS_SESSION_ID"] = str(session_id)
            env["RANK"] = str(op.rank)

        @classmethod
        def _build_locals(cls, ctx, op):
            logger.debug("Start to create mars session.")
            sess = ctx.get_current_session().as_default()

            return dict(session=sess)

        @classmethod
        def execute(cls, ctx, op):
            if op.merge:
                return super().execute(ctx, op)

            old_env = os.environ.copy()
            cls._set_envs(ctx, op)

            try:
                if op.mode == "spawn":
                    cls._execute_with_subprocess(op)
                elif op.mode == "exec":
                    cls._execute_with_exec(op, local=cls._build_locals(ctx, op))
                else:
                    raise TypeError("Unsupported mode {}".format(op.mode))

                if op.rank == 0:
                    ctx[op.outputs[0].key] = {"status": "ok"}
                else:
                    ctx[op.outputs[0].key] = {}
            finally:
                os.environ = old_env


def _build_locals_in_odps(cls, ctx, op):
    from odps import ODPS

    o = ODPS.from_environments()
    sess = ctx.get_current_session().as_default()
    return dict(o=o, session=sess, odps=o)


RunScript._build_locals = classmethod(_build_locals_in_odps)


def run_script(
    script,
    n_workers=1,
    mode="exec",
    command_argv=None,
    odps_params=None,
    session=None,
    run_kwargs=None,
):
    if hasattr(script, "read"):
        code = script.read()
    else:
        with open(os.path.abspath(script), "rb") as f:
            code = f.read()
    if mode not in ["exec", "spawn"]:
        raise TypeError("Unsupported mode {}".format(mode))

    op = RunScript(
        code=to_binary(code), mode=mode, world_size=n_workers, command_args=command_argv
    )
    op.extra_params["project"] = odps_params["project"]
    op.extra_params["endpoint"] = (
        os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
    )
    return op().execute(session=session, **(run_kwargs or {})).fetch(session=session)



# encoding: utf-8
# copyright 1999-2022 alibaba group holding ltd.
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#      http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

import types


def get_field_name_by_role(df, role):
    fields = [f for f in df._ml_fields if role in f.role]
    if not fields:
        raise ValueError('Input df does not contain a field with role %s.' % role.name)
    return fields[0].name


class FuncArgs(Exception):
    def __init__(self, *args, **kwargs):
        self._args = kwargs.pop('args', ())
        self._kwargs = kwargs.pop('kwargs', {})
        super(FuncArgs, self).__init__(*args, **kwargs)


def replace_function_globals(fun, repl_dict):
    glob = fun.__globals__.copy()
    glob.update(repl_dict)
    new_fun = types.FunctionType(fun.__code__, glob)
    new_fun.__name__ = fun.__name__
    new_fun.__defaults__ = fun.__defaults__
    if hasattr(fun, '__kwdefaults__'):
        new_fun.__kwdefaults__ = fun.__kwdefaults__
    return new_fun


def fetch_call_args(fun_name, callee, *args, **kwargs):
    def _raiser(*args, **kwargs):
        raise FuncArgs(args=args, kwargs=kwargs)

    new_callee = replace_function_globals(callee, {fun_name: _raiser})
    try:
        new_callee(*args, **kwargs)
    except FuncArgs as ex:
        return ex._args, ex._kwargs
    return (), {}


def metrics_result(runner_fun, fun_name=None):
    fun_name = fun_name or runner_fun.__name__

    def _decorator(fun):
        def _decorated(*args, **kwargs):
            execute_now = kwargs.pop('execute_now', True)
            if execute_now:
                return fun(*args, **kwargs)
            else:
                def result_callback(ret):
                    new_fun = replace_function_globals(fun, {fun_name: lambda *args, **kw: ret})
                    return new_fun(*args, **kwargs)

                node_args, node_kw = fetch_call_args(fun_name, fun, *args, **kwargs)
                node_kw.update(dict(execute_now=False, result_callback=result_callback))
                return runner_fun(*node_args, **node_kw)

        _decorated.__name__ = fun.__name__
        _decorated.__doc__ = fun.__doc__
        return _decorated
    return _decorator

_metrics_fallback_detected = False


def detect_metrics_fallback(df):
    global _metrics_fallback_detected
    from ... import ODPS, errors, options

    if _metrics_fallback_detected or options.ml.use_old_metrics:
        return

    _metrics_fallback_detected = True
    odps = ODPS.from_global()
    for src in df.data_source():
        if hasattr(src, 'project'):
            odps = src.project.odps

    try:
        odps.execute_xflow('MultiClassEvaluation', xflow_project=options.ml.xflow_project)
    except errors.ODPSError as ex:
        if 'xflow not found' in str(ex):
            options.ml.use_old_metrics = True



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *


cdef class CDecoder:
    cdef size_t _pos
    cdef object _stream
    cdef bytes _buffer
    cdef char* _begin
    cdef char* _end
    cdef bint _is_source_eof

    cdef size_t position(self) nogil
    cdef int32_t read_field_number(self) except? -1 nogil
    cdef read_field_number_and_wire_type(self)
    cdef int32_t read_sint32(self) except? -1 nogil
    cdef uint32_t read_uint32(self) except? 0xffffffff nogil
    cdef int64_t read_sint64(self) except? -1 nogil
    cdef uint64_t read_uint64(self) except? 0xffffffff nogil
    cdef bint read_bool(self) except? False nogil
    cdef double read_double(self) except? -1.0 nogil
    cdef float read_float(self) except? -1.0 nogil
    cdef bytes read_string(self)
    cdef int _load_next_buffer(self) except -1 with gil


cdef class Decoder:
    cdef CDecoder _decoder



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
from libc.stdint cimport *
from cpython.datetime cimport import_datetime

from .. import types
from ..src.types_c cimport BaseRecord
from ..src.utils_c cimport CMillisecondsConverter

cdef:
    int64_t BOOL_TYPE_ID = types.boolean._type_id
    int64_t DATETIME_TYPE_ID = types.datetime._type_id
    int64_t DATE_TYPE_ID = types.date._type_id
    int64_t STRING_TYPE_ID = types.string._type_id
    int64_t FLOAT_TYPE_ID = types.float_._type_id
    int64_t DOUBLE_TYPE_ID = types.double._type_id
    int64_t BIGINT_TYPE_ID = types.bigint._type_id
    int64_t BINARY_TYPE_ID = types.binary._type_id
    int64_t TIMESTAMP_TYPE_ID = types.timestamp._type_id
    int64_t TIMESTAMP_NTZ_TYPE_ID = types.timestamp_ntz._type_id
    int64_t INTERVAL_DAY_TIME_TYPE_ID = types.interval_day_time._type_id
    int64_t INTERVAL_YEAR_MONTH_TYPE_ID = types.interval_year_month._type_id
    int64_t DECIMAL_TYPE_ID = types.Decimal._type_id
    int64_t JSON_TYPE_ID = types.Json._type_id

import_datetime()


cdef class AbstractHasher:
    cdef int32_t c_hash_bigint(self, int64_t val) nogil:
        return 0

    def hash_bigint(self, int64_t val):
        return self.c_hash_bigint(val)

    cdef int32_t c_hash_float(self, float val) nogil:
        cdef int32_t *ptr = <int32_t *>&val
        return self.c_hash_bigint(ptr[0])

    def hash_float(self, float val):
        return self.c_hash_float(val)

    cdef int32_t c_hash_double(self, double val) nogil:
        cdef int64_t *ptr = <int64_t *>&val
        return self.c_hash_bigint(ptr[0])

    def hash_double(self, double val):
        return self.c_hash_double(val)

    cdef int32_t c_hash_bool(self, bint val) nogil:
        return 0

    def hash_bool(self, bint val):
        return self.c_hash_bool(val)

    cdef int32_t c_hash_string(self, char *ptr, size_t size) nogil:
        return 0

    def hash_string(self, val):
        cdef bytes bval
        if isinstance(val, unicode):
            bval = (<unicode>val).encode()
        else:
            bval = val
        return self.c_hash_string(bval, len(bval))


cdef class DefaultHasher(AbstractHasher):
    cdef int32_t c_hash_bigint(self, int64_t val) nogil:
        val = (~val) + (val << 18)
        val ^= val >> 31
        val *= <long>21
        val ^= val >> 11
        val += val << 6
        val ^= val >> 22
        return <int32_t>val

    cdef int32_t  c_hash_bool(self, bint val) nogil:
        # it is a magic number
        if val:
            return 0x172ba9c7
        else:
            return -0x3a59cb12

    cdef int32_t  c_hash_string(self, char *ptr, size_t size) nogil:
        cdef int32_t hash_val = 0
        for i in range(size):
            hash_val += ptr[i]
            hash_val += hash_val << 10
            hash_val ^= hash_val >> 6
        hash_val += hash_val << 3
        hash_val ^= hash_val >> 11
        hash_val += hash_val << 15
        return hash_val


cdef class LegacyHasher(AbstractHasher):
    cdef int32_t  c_hash_bigint(self, int64_t val) nogil:
        return (val >> 32) ^ val

    cdef int32_t  c_hash_bool(self, bint val) nogil:
        # it is a magic number
        if val:
            return 0x172ba9c7
        else:
            return -0x3a59cb12

    cdef int32_t  c_hash_string(self, char *ptr, size_t size) nogil:
        cdef int32_t hash_val = 0
        for i in range(size):
            hash_val = hash_val * 31 + ptr[i]
        return hash_val


cpdef AbstractHasher get_hasher(hasher_type):
    if hasher_type == "legacy":
        return LegacyHasher()
    elif hasher_type == "default":
        return DefaultHasher()
    else:
        raise ValueError("Hasher type %s not supported" % hasher_type)


def _hash_timestamp(hasher, x):
    seconds = int(x.timestamp())
    nanos = x.microsecond * 1000 + x.nanosecond
    return hasher.hash_bigint((seconds << 30) | nanos)


def _hash_timedelta(hasher, x):
    seconds = int(x.total_seconds())
    nanos = x.microsecond * 1000 + x.nanosecond
    return hasher.hash_bigint((seconds << 30) | nanos)


cdef class RecordHasher:
    def __init__(self, schema, hasher_type, hash_keys):
        cdef int type_id, i
        cdef int key_count = len(hash_keys)
        cdef set hash_keys_set = set(hash_keys)

        self._schema_snapshot = schema.build_snapshot()
        self._hasher = get_hasher(hasher_type)
        self._mills_converter = CMillisecondsConverter()

        self._col_ids.reserve(key_count)
        self._idx_to_hash_fun.reserve(key_count)
        for i in range(self._schema_snapshot._col_count):
            if self._schema_snapshot._columns[i].name not in hash_keys_set:
                continue

            self._col_ids.push_back(i)
            type_id = self._schema_snapshot._col_type_ids[i]
            if type_id == BIGINT_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_bigint)
            elif type_id == FLOAT_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_float)
            elif type_id == DOUBLE_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_double)
            elif type_id == STRING_TYPE_ID or type_id == BINARY_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_string)
            elif type_id == BOOL_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_bool)
            elif type_id == DATE_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_date)
            elif type_id == DATETIME_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_datetime)
            elif type_id == TIMESTAMP_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_timestamp)
            elif type_id == INTERVAL_DAY_TIME_TYPE_ID:
                self._idx_to_hash_fun.push_back(self._hash_timedelta)
            else:
                raise TypeError(
                    "Hash for type %s not supported" % self._schema_snapshot._col_types[i]
                )

    cdef int32_t _hash_bigint(self, object value) except? -1:
        return self._hasher.c_hash_bigint(value)

    cdef int32_t _hash_float(self, object value) except? -1:
        return self._hasher.c_hash_float(value)

    cdef int32_t _hash_double(self, object value) except? -1:
        return self._hasher.c_hash_double(value)

    cdef int32_t _hash_bool(self, object value) except? -1:
        return self._hasher.c_hash_bool(value)

    cdef int32_t _hash_string(self, object value) except? -1:
        cdef bytes bval
        if isinstance(value, unicode):
            bval = (<unicode>value).encode()
        else:
            bval = value
        return self._hasher.c_hash_string(bval, len(bval))

    cdef int32_t _hash_date(self, object value) except? -1:
        return self._hasher.c_hash_bigint(int(calendar.timegm(value.timetuple())))

    cdef int32_t _hash_datetime(self, object value) except? -1:
        return self._hasher.c_hash_bigint(self._mills_converter.to_milliseconds(value))

    cdef int32_t _hash_timestamp(self, object value) except? -1:
        cdef int64_t seconds = int(
            self._mills_converter.to_milliseconds(value.to_pydatetime()) / 1000
        )
        cdef int64_t nanos = value.microsecond * 1000 + value.nanosecond
        return self._hasher.hash_bigint((seconds << 30) | nanos)

    cdef int32_t _hash_timedelta(self, object value) except? -1:
        cdef int64_t seconds = int(value.total_seconds())
        cdef int64_t nanos = value.microseconds * 1000 + value.nanoseconds
        return self._hasher.hash_bigint((seconds << 30) | nanos)

    cpdef int32_t hash(self, BaseRecord record):
        cdef int i
        cdef int32_t hash_sum = 0

        for i in range(self._col_ids.size()):
            if record._c_values[<int>self._col_ids[i]] is None:
                continue
            hash_sum += self._idx_to_hash_fun[i](self, record._c_values[i])
        return hash_sum ^ (hash_sum >> 8)


cpdef int32_t hash_value(hasher_type, data_type, value):
    """Simple hash function for test purpose"""
    cdef RecordHasher rec_hasher

    from ..types import Column, OdpsSchema
    from ..models import Record

    schema = OdpsSchema([Column("col", data_type)])
    record = Record(schema=schema, values=[value])

    rec_hasher = RecordHasher(schema, hasher_type, ["col"])
    return rec_hasher._idx_to_hash_fun[0](rec_hasher, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from mars.utils import to_binary


logger = logging.getLogger(__name__)

from mars.remote.run_script import RunScript


def _build_locals_in_odps(cls, ctx, op):
    from odps import ODPS

    sess = ctx.get_current_session().as_default()
    local = {"session": sess}
    if op.data is not None:
        local.update(op.data)
    o = ODPS.from_environments()
    return dict(o=o, session=sess, odps=o)


def execute_with_odps_context(f):
    def wrapper(ctx, op):
        from .cupid_service import CupidServiceClient
        from mars.utils import to_str

        if "CUPID_SERVICE_SOCKET" not in os.environ:
            f(ctx, op)
        else:
            old_envs = os.environ.copy()
            try:
                env = os.environ

                logger.debug("Get bearer token from cupid.")
                bearer_token = CupidServiceClient().get_bearer_token()
                env["ODPS_BEARER_TOKEN"] = to_str(bearer_token)
                if "endpoint" in op.extra_params:
                    env["ODPS_ENDPOINT"] = os.environ.get(
                        "ODPS_RUNTIME_ENDPOINT"
                    ) or str(op.extra_params["endpoint"])
                if ("project" in op.extra_params) and ("ODPS_PROJECT_NAME" not in env):
                    env["ODPS_PROJECT_NAME"] = str(op.extra_params["project"])
                f(ctx, op)
                for out in op.outputs:
                    if ctx[out.key] is None:
                        ctx[out.key] = {"status": "OK"}
            finally:
                os.environ = old_envs

    return wrapper


RunScript._build_locals = classmethod(_build_locals_in_odps)


def run_script(
    script,
    data=None,
    n_workers=1,
    mode="exec",
    command_argv=None,
    retry_when_fail=False,
    odps_params=None,
    session=None,
    run_kwargs=None,
):
    from mars.remote.run_script import _extract_inputs

    if hasattr(script, "read"):
        code = script.read()
    else:
        with open(os.path.abspath(script), "rb") as f:
            code = f.read()
    inputs = _extract_inputs(data)
    op = RunScript(
        data=data,
        code=to_binary(code),
        world_size=n_workers,
        retry_when_fail=retry_when_fail,
        command_args=command_argv,
    )

    op.extra_params["project"] = odps_params["project"]
    op.extra_params["endpoint"] = (
        os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
    )
    return (
        op(inputs).execute(session=session, **(run_kwargs or {})).fetch(session=session)
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import sqlalchemy
    from sqlalchemy import create_engine, Column, Integer, MetaData

    from .ext import *

    has_sqlalchemy = True
except ImportError:
    has_sqlalchemy = False

from ....models import Table
from ....config import options
from ....types import Decimal
from ... import types as df_types
from ...expr.reduction import Max, GroupedMax, Min, GroupedMin
from ..sqlalchemy.compiler import SQLAlchemyCompiler
from ..sqlalchemy.types import df_schema_to_sqlalchemy_columns
from .models import SeahawksTable


_url_to_sqlalchemy_engine_and_metadatas = {}


class SeahawksCompiler(SQLAlchemyCompiler):

    def __init__(self, expr_dag, odps):
        super(SeahawksCompiler, self).__init__(expr_dag)
        self._odps = odps

        seahawks_url = odps._seahawks_url or options.seahawks_url
        if seahawks_url in _url_to_sqlalchemy_engine_and_metadatas:
            self._sqlalchemy_engine, self._metadata = \
                _url_to_sqlalchemy_engine_and_metadatas[seahawks_url]
        else:
            _sqlalchemy_engine = self._sqlalchemy_engine = \
                create_engine(seahawks_url, isolation_level='AUTOCOMMIT')
            _sqlalchemy_engine.dialect.identifier_preparer.initial_quote = ''
            _sqlalchemy_engine.dialect.identifier_preparer.final_quote = ''
            self._metadata = MetaData(bind=_sqlalchemy_engine)
            _url_to_sqlalchemy_engine_and_metadatas[seahawks_url] = \
                _sqlalchemy_engine, self._metadata

    def _mapping_odps_table(self, t, df_schema):
        if isinstance(t, SeahawksTable):  # from cache, actually heap table
            table_name = t.name
        else:
            table_name = 'odps.{0}.{1}'.format(t.project.name, t.name)
        columns = df_schema_to_sqlalchemy_columns(df_schema)
        return sqlalchemy.Table(table_name, self._metadata, *columns, extend_existing=True)

    def visit_source_collection(self, expr):
        table = next(expr.data_source())

        if isinstance(table, Table):
            if any(isinstance(col.type, Decimal) for col in table.table_schema.columns):
                # FIXME: decimal and datetime are not supported by seahawks by now
                raise NotImplementedError
            table = self._mapping_odps_table(table, expr.schema)

        self._add(expr, table.alias(self._new_alias()))

    def visit_reduction(self, expr):
        if isinstance(expr, (Max, Min, GroupedMax, GroupedMin)) and \
                expr._input.dtype == df_types.string:
            # string max or min act different
            raise NotImplementedError

        return super(SeahawksCompiler, self).visit_reduction(expr)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import re

if sys.version_info[0] <= 2:
    string_types = basestring
else:
    string_types = str

_MD_PATTERNS = [
    re.compile(r'^(?P<month>(\d+|-\d+))$'),
    re.compile(r'^(?P<year>(\d+|-\d+))\s*-\s*(?P<month>(\d+|-\d+))$'),
    re.compile(r'^(?P<year>(\d+|-\d+))\s*(years|year)\s*(?P<month>(\d+|-\d+))\s*(months|month)$'),
    re.compile(r'^(?P<year>(\d+|-\d+))\s*(years|year)$'),
    re.compile(r'^(?P<month>(\d+|-\d+))\s*(months|month)$'),
]


class Monthdelta(object):
    __slots__ = '_total_months',

    def __init__(self, years=None, months=None):
        if isinstance(years, string_types):
            years, months = self._parse(years)
        elif months is None:
            years, months = months, years
        years = years or 0
        months = months or 0
        self._total_months = int(years * 12 + months)

    @classmethod
    def _parse(cls, s):
        s = s.strip()
        for p in _MD_PATTERNS:
            match = p.match(s)
            if not match:
                continue
            match = match.groupdict()
            return int(match.get('year', 0)), int(match.get('month', 0))
        raise ValueError("Cannot parse '%s' for %s." % (s, cls.__name__))

    def total_months(self):
        return self._total_months

    @property
    def months(self):
        return self._total_months % 12

    @property
    def years(self):
        return self._total_months // 12

    def __getstate__(self):
        return dict(_total_months=self._total_months)

    def __setstate__(self, state):
        self._total_months = state['_total_months']

    def __int__(self):
        return self.total_months()

    if sys.version_info[0] <= 2:
        def __long__(self):
            return self.total_months()

    def __str__(self):
        years = self.years
        months = self.months
        m = lambda v: 's' if abs(v) > 1 else ''
        if years == 0 and months == 0:
            return '0'
        elif years == 0:
            return '%d month%s' % (months, m(months))
        elif months == 0:
            return '%d year%s' % (years, m(years))
        else:
            return '%d year%s %d month%s' % (years, m(years), months, m(months))

    def __repr__(self):
        return "%s('%s')" % (type(self).__name__, str(self))

    def __eq__(self, other):
        if not isinstance(other, Monthdelta):
            return False
        return self._total_months == other._total_months

    def __ne__(self, other):
        return not (self == other)

    def __gt__(self, other):
        try:
            return self._total_months > other._total_months
        except AttributeError:
            raise TypeError("unsupported operand type(s) for >: '%s' and '%s'" %
                            (type(self).__name__, type(other).__name__))

    def __ge__(self, other):
        try:
            return self._total_months >= other._total_months
        except AttributeError:
            raise TypeError("unsupported operand type(s) for >=: '%s' and '%s'" %
                            (type(self).__name__, type(other).__name__))

    def __lt__(self, other):
        try:
            return self._total_months < other._total_months
        except AttributeError:
            raise TypeError("unsupported operand type(s) for <: '%s' and '%s'" %
                            (type(self).__name__, type(other).__name__))

    def __le__(self, other):
        try:
            return self._total_months <= other._total_months
        except AttributeError:
            raise TypeError("unsupported operand type(s) for <=: '%s' and '%s'" %
                            (type(self).__name__, type(other).__name__))

    def __abs__(self):
        return Monthdelta(months=abs(self._total_months))

    def __neg__(self):
        return Monthdelta(months=-self._total_months)

    def _add(self, other):
        if isinstance(other, Monthdelta):
            return Monthdelta(months=self._total_months + other._total_months)

        new_year = other.year + self.years
        new_month = other.month + self.months
        if new_month < 0:
            new_month += 12
            new_year -= 1
        new_year += new_month // 12
        new_month = new_month % 12
        return other.replace(year=new_year, month=new_month)

    def __add__(self, other):
        try:
            return self._add(other)
        except AttributeError:
            raise TypeError("unsupported operand type(s) for +: '%s' and '%s'" %
                            (type(self).__name__, type(other).__name__))

    def __radd__(self, other):
        try:
            return self._add(other)
        except AttributeError:
            raise TypeError("unsupported operand type(s) for +: '%s' and '%s'" %
                            (type(other).__name__, type(self).__name__))

    def __sub__(self, other):
        if isinstance(other, Monthdelta):
            return Monthdelta(months=self._total_months - other._total_months)
        else:
            raise TypeError("unsupported operand type(s) for +: '%s' and '%s'" %
                            (type(other).__name__, type(self).__name__))

    def __rsub__(self, other):
        try:
            return (-self)._add(other)
        except AttributeError:
            raise TypeError("unsupported operand type(s) for -: '%s' and '%s'" %
                            (type(other).__name__, type(self).__name__))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

try:
    import pandas as pd
    has_pandas = True
except ImportError:
    has_pandas = False

try:
    import sqlalchemy
    has_sqlalchemy = True
except ImportError:
    has_sqlalchemy = False

from ...models import Table
from ...models.ml import OfflineModel
from ..expr.expressions import CollectionExpr
from ..expr.core import ExprDictionary
from .odpssql.models import MemCacheReference
from .seahawks.models import SeahawksTable
from .core import EngineTypes as Engines
from .errors import NoBackendFound
from .utils import fetch_data_source_size
from ...config import options


def available_engines(sources):
    from ...ml.models import TablesModelObject
    engines = set()

    for src in sources:
        if isinstance(src, SeahawksTable):
            engines.add(Engines.SEAHAWKS)
        elif isinstance(src, (Table, MemCacheReference)):
            engines.add(Engines.ODPS)
        elif isinstance(src, (OfflineModel, TablesModelObject)):
            engines.add(Engines.ALGO)
        elif has_pandas and isinstance(src, pd.DataFrame):
            engines.add(Engines.PANDAS)
        elif has_sqlalchemy and isinstance(src, sqlalchemy.Table):
            engines.add(Engines.SQLALCHEMY)
        else:
            raise NoBackendFound('No backend found for data source: %s' % src)

    return engines


class EngineSelecter(object):
    def __init__(self):
        self._node_engines = ExprDictionary()
        self.force_odps = False

    def _choose_backend(self, node):
        if node in self._node_engines:
            return self._node_engines[node]

        available_engine_size = len(Engines)

        engines = set()
        children = node.children()
        if len(children) == 0:
            if not isinstance(node, CollectionExpr):
                self._node_engines[node] = None
                return
            engines.update(available_engines(node._data_source()))
        else:
            for n in children:
                e = self._choose_backend(n)
                if e and e not in engines:
                    engines.add(e)
                if len(engines) == available_engine_size:
                    break

        if len(engines) == 1:
            self._node_engines[node] = engines.pop()
        else:
            self._node_engines[node] = self._choose(node)

        return self._node_engines[node]

    def has_diff_data_sources(self, node, no_cache=False):
        if no_cache:
            return len(available_engines(node.data_source())) > 1

        children = node.children()
        if len(children) == 0:
            return False
        return len(set(filter(lambda x: x is not None,
                              (self._choose_backend(c) for c in children)))) > 1

    def _has_data_source(self, node, src):
        children = node.children()
        if len(children) == 0:
            return src == self._choose_backend(node)
        return src in set(self._choose_backend(c) for c in node.children())

    def has_pandas_data_source(self, node):
        return self._has_data_source(node, Engines.PANDAS)

    def has_odps_data_source(self, node):
        return self._has_data_source(node, Engines.ODPS)

    def _choose(self, node):
        # Now if we do an operation like join
        # which comes from different data sources
        # we just use ODPS to caculate.
        # In the future, we will choose according to the scale of data

        return Engines.ODPS

    def _choose_odps_backend(self, node_dag):
        node = node_dag.root

        if not has_sqlalchemy or not options.seahawks_url:
            return Engines.ODPS

        sizes = 0
        for n in node.traverse(top_down=True, unique=True):
            for ds in n._data_source():
                if isinstance(ds, MemCacheReference):
                    return Engines.ODPS  # if use mem cache, directly return ODPSSQL backend
                if not isinstance(ds, Table) or isinstance(ds, SeahawksTable):
                    continue
                if isinstance(ds, Table) and n._deps is not None:
                    # the cached collection, the table has not been calculated yet
                    continue
                size = fetch_data_source_size(node_dag, n, ds)
                if size is None:
                    return Engines.ODPS
                sizes += size

        if sizes <= options.df.seahawks.max_size:
            return Engines.SEAHAWKS

        return Engines.ODPS

    def select(self, expr_dag):
        node = expr_dag.root

        if getattr(node, '_algo', None) is not None:
            return Engines.ALGO

        if not self.has_diff_data_sources(node):
            engine = self._choose_backend(node)
            if engine == Engines.ODPS and not self.force_odps:
                return self._choose_odps_backend(expr_dag)
            return engine

        engines = available_engines(node.data_source())
        if len(engines) > 1 or (len(engines) == 1 and list(engines)[0] == Engines.ODPS):
            if not self.force_odps:
                return self._choose_odps_backend(expr_dag)
            else:
                return Engines.ODPS
        return engines.pop()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from collections import namedtuple

from .. import serializers, utils, types, errors, compat
from ..compat import Enum, six
from ..config import options
from .core import LazyLoad
from .cache import cache, cache_parent

_RESOURCE_SPLITTER = '/resources/'
_SCHEMA_SPLITTER = '/schemas/'

if sys.version_info[0] < 3:
    _StringIOType = type(compat.StringIO())
else:
    _StringIOType = (compat.StringIO, compat.BytesIO)


class Resource(LazyLoad):
    """
    Resource is useful when writing UDF or MapReduce. This is an abstract class.

    Basically, resource can be either a file resource or a table resource.
    File resource can be ``file``, ``py``, ``jar``, ``archive`` in details.

    .. seealso:: :class:`odps.models.FileResource`, :class:`odps.models.PyResource`,
                 :class:`odps.models.JarResource`, :class:`odps.models.ArchiveResource`,
                 :class:`odps.models.TableResource`
    """

    __slots__ = (
        'content_md5', 'is_temp_resource', 'volume_path', '_type_indicator'
    )

    class Type(Enum):
        FILE = 'FILE'
        JAR = 'JAR'
        PY = 'PY'
        ARCHIVE = 'ARCHIVE'
        TABLE = 'TABLE'
        VOLUMEFILE = 'VOLUMEFILE'
        VOLUMEARCHIVE = 'VOLUMEARCHIVE'
        UNKOWN = 'UNKOWN'

    _type_indicator = 'type'

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    comment = serializers.XMLNodeField('Comment')
    type = serializers.XMLNodeField('ResourceType', parse_callback=lambda t: Resource.Type(t.upper()))
    creation_time = serializers.XMLNodeField('CreationTime', parse_callback=utils.parse_rfc822)
    last_modified_time = serializers.XMLNodeField('LastModifiedTime', parse_callback=utils.parse_rfc822)
    last_updator = serializers.XMLNodeField('LastUpdator')
    size = serializers.XMLNodeField('ResourceSize', parse_callback=int)
    source_table_name = serializers.XMLNodeField('TableName')

    @classmethod
    def _get_cls(cls, typo):
        if typo is None:
            return cls

        if isinstance(typo, six.string_types):
            typo = Resource.Type(typo.upper())

        clz = lambda name: globals()[name]
        if typo == Resource.Type.FILE:
            return clz('FileResource')
        elif typo == Resource.Type.JAR:
            return clz('JarResource')
        elif typo == Resource.Type.PY:
            return clz('PyResource')
        elif typo == Resource.Type.ARCHIVE:
            return clz('ArchiveResource')
        elif typo == Resource.Type.TABLE:
            return clz('TableResource')
        elif typo == Resource.Type.VOLUMEARCHIVE:
            return clz('VolumeArchiveResource')
        elif typo == Resource.Type.VOLUMEFILE:
            return clz('VolumeFileResource')
        else:
            return cls

    def create(self, overwrite=False, **kw):
        raise NotImplementedError

    @staticmethod
    def _filter_cache(_, **kwargs):
        return kwargs.get('type') is not None and kwargs['type'] != Resource.Type.UNKOWN

    @cache
    def __new__(cls, *args, **kwargs):
        typo = kwargs.get('type')
        if typo is not None or (cls != Resource and issubclass(cls, Resource)):
            return object.__new__(cls._get_cls(typo))

        kwargs['type'] = Resource.Type.UNKOWN
        obj = Resource(**kwargs)
        obj.reload()
        return Resource(**obj.extract())

    def __init__(self, **kwargs):
        typo = kwargs.get('type')
        if isinstance(typo, six.string_types):
            kwargs['type'] = Resource.Type(typo.upper())
        super(Resource, self).__init__(**kwargs)

    @classmethod
    def build_full_resource_name(cls, name, project_name, schema_name=None):
        if project_name is None:
            return name
        elif schema_name is not None:
            return project_name + _SCHEMA_SPLITTER + schema_name + _RESOURCE_SPLITTER + name
        else:
            return project_name + _RESOURCE_SPLITTER + name

    @classmethod
    def split_resource_name(cls, name):
        project_name, schema_name = None, None
        if _RESOURCE_SPLITTER in name:
            project_schema_name, name = name.split(_RESOURCE_SPLITTER, 1)

            if _SCHEMA_SPLITTER not in project_schema_name:
                project_name, schema_name = project_schema_name, None
            else:
                project_name, schema_name = project_schema_name.split(_SCHEMA_SPLITTER, 1)
        return project_name, schema_name, name

    @property
    def full_resource_name(self):
        schema_name = self._get_schema_name()
        return self.build_full_resource_name(self.name, self.project.name, schema_name)

    def reload(self):
        params = {}
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name

        url = self.resource()
        resp = self._client.get(url, action='meta', params=params)

        self.owner = resp.headers.get('x-odps-owner')
        resource_type = resp.headers.get('x-odps-resource-type')
        self.type = Resource.Type(resource_type.upper())
        self.comment = resp.headers.get('x-odps-comment')
        self.last_updator = resp.headers.get('x-odps-updator')

        size = resp.headers.get('x-odps-resource-size')
        self.size = None if size is None else int(size)

        self.creation_time = utils.parse_rfc822(
            resp.headers.get('x-odps-creation-time'))
        self.last_modified_time = utils.parse_rfc822(
            resp.headers.get('Last-Modified'))

        is_temp_resource_header = resp.headers.get("x-odps-resource-istemp") or ""
        self.is_temp_resource = is_temp_resource_header.lower() == "true"

        self.source_table_name = resp.headers.get('x-odps-copy-table-source')
        self.volume_path = resp.headers.get('x-odps-copy-file-source')
        self.content_md5 = resp.headers.get('Content-MD5')

        self._loaded = True

    def _reload_size(self):
        params = {}
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name

        url = self.resource()
        resp = self._client.get(url, action='meta', params=params)

        size = resp.headers.get('x-odps-resource-size')
        self.size = None if size is None else int(size)

    def update(self, **kw):
        raise NotImplementedError

    def drop(self):
        return self.parent.delete(self)


@cache_parent
class FileResource(Resource):
    """
    File resource represents for a file.

    Use ``open`` method to open this resource as a file-like object.
    """

    __slots__ = ('_fp', 'is_part_resource', 'merge_total_bytes')

    class Mode(Enum):
        READ = 'r'
        WRITE = 'w'
        APPEND = 'a'
        READWRITE = 'r+'
        TRUNCEREADWRITE = 'w+'
        APPENDREADWRITE = 'a+'

    def create(self, overwrite=False, **kw):
        file_obj = kw.pop('file_obj', kw.pop('fileobj', None))
        is_part_resource = self._getattr("is_part_resource")
        is_merge_resource = self._getattr('merge_total_bytes') is not None

        if file_obj is None:
            raise ValueError('parameter `file_obj` cannot be None, either string or file-like object')
        if isinstance(file_obj, six.text_type):
            file_obj = file_obj.encode('utf-8')

        if (
            options.upload_resource_in_chunks
            and not is_merge_resource
            and not is_part_resource
            and self._get_file_size(file_obj) > options.resource_chunk_size
        ):
            self._upload_with_stream(file_obj, overwrite)
            return self

        if self.name is None or len(self.name.strip()) == 0:
            raise errors.ODPSError('File Resource Name should not empty.')

        method = self._client.post if not overwrite else self._client.put
        url = self.parent.resource() if not overwrite else self.resource()

        headers = {
            'Content-Type': 'application/octet-stream',
            'Content-Disposition': 'attachment;filename=%s' % self.name,
            'x-odps-resource-type': self.type.value.lower(),
            'x-odps-resource-name': self.name,
        }
        params = {}

        if self._getattr('comment') is not None:
            headers['x-odps-comment'] = self.comment
        if self._getattr('is_temp_resource'):
            headers['x-odps-resource-istemp'] = 'true' if self.is_temp_resource else 'false'
        if is_merge_resource:
            headers['x-odps-resource-merge-total-bytes'] = str(self.merge_total_bytes)
            params["rOpMerge"] = "true"
        if is_part_resource:
            params["rIsPart"] = "true"

        if not isinstance(file_obj, (six.string_types, six.binary_type)):
            file_obj.seek(0)
            content = file_obj.read()
        else:
            content = file_obj

        self.size = len(content)
        method(
            url, content, headers=headers, params=params, curr_schema=self._get_schema_name()
        )

        if overwrite:
            self.reload()
        return self

    def __init__(self, **kw):
        super(FileResource, self).__init__(**kw)
        self.type = Resource.Type.FILE

        self._fp = None

    @staticmethod
    def _get_file_size(file_obj):
        if isinstance(file_obj, six.binary_type):
            return len(file_obj)
        elif isinstance(file_obj, _StringIOType):
            pos = file_obj.tell()
            file_obj.seek(0, os.SEEK_END)
            size = file_obj.tell()
            file_obj.seek(pos, os.SEEK_SET)
            return size
        elif hasattr(file_obj, "name"):
            fn = file_obj.name
            if fn and os.path.exists(fn):
                return os.path.getsize(fn)
        return None

    def _upload_with_stream(self, file_obj, overwrite=True):
        chunk_size = options.resource_chunk_size
        with self.open("wb", stream=True, overwrite=overwrite) as res_file:
            while True:
                buf = file_obj.read(chunk_size)
                if not buf:
                    break
                res_file.write(buf)

    def _is_create(self):
        if self._loaded:
            return False
        try:
            self._reload_size()
            return False
        except errors.NoSuchObject:
            return True

    @property
    def _size(self):
        return self._fp.size if self._fp else self.size

    @property
    def opened(self):
        return self._fp is not None

    @property
    def mode(self):
        return self._fp.mode

    def open(self, mode='r', encoding='utf-8', stream=False, overwrite=None):
        """
        The argument ``mode`` stands for the open mode for this file resource.
        It can be binary mode if the 'b' is inside. For instance,
        'rb' means opening the resource as read binary mode
        while 'r+b' means opening the resource as read+write binary mode.
        This is most import when the file is actually binary such as tar or jpeg file,
        so be aware of opening this file as a correct mode.

        Basically, the text mode can be 'r', 'w', 'a', 'r+', 'w+', 'a+'
        just like the builtin python ``open`` method.

        * ``r`` means read only
        * ``w`` means write only, the file will be truncated when opening
        * ``a`` means append only
        * ``r+`` means read+write without constraint
        * ``w+`` will truncate first then opening into read+write
        * ``a+`` can read+write, however the written content can only be appended to the end

        :param mode: the mode of opening file, described as above
        :param encoding: utf-8 as default
        :param stream: open in stream mode
        :param overwrite: if True, will overwrite existing resource. True by default.
        :return: file-like object

        :Example:

        >>> with resource.open('r') as fp:
        >>>     fp.read(1)  # read one unicode character
        >>>     fp.write('test')  # wrong, cannot write under read mode
        >>>
        >>> with resource.open('wb') as fp:
        >>>     fp.readlines()  # wrong, cannot read under write mode
        >>>     fp.write('hello world')  # write bytes
        >>>
        >>> with resource.open('test_resource', 'r+') as fp:  # open as read-write mode
        >>>     fp.seek(5)
        >>>     fp.truncate()
        >>>     fp.flush()
        """

        from .resourcefile import LocalResourceFile, StreamResourceFile

        if stream:
            res_file_type = StreamResourceFile
        else:
            res_file_type = LocalResourceFile

        self._fp = res_file_type(self, mode, encoding, overwrite=overwrite)
        return self

    def _check_read(self):
        if not self.opened:
            raise IOError('I/O operation on non-open resource')
        if self.mode in (FileResource.Mode.WRITE, FileResource.Mode.APPEND):
            raise IOError('Resource not open for reading')

    def read(self, size=-1):
        """
        Read the file resource, read all as default.

        :param size: unicode or byte length depends on text mode or binary mode.
        :return: unicode or bytes depends on text mode or binary mode
        :rtype: str or unicode(Py2), bytes or str(Py3)
        """

        self._check_read()
        return self._fp.read(size)

    def readline(self, size=-1):
        """
        Read a single line.

        :param size: If the size argument is present and non-negative,
                     it is a maximum byte count (including the trailing newline)
                     and an incomplete line may be returned.
                     When size is not 0,
                     an empty string is returned only when EOF is encountered immediately
        :return: unicode or bytes depends on text mode or binary mode
        :rtype: str or unicode(Py2), bytes or str(Py3)
        """

        self._check_read()
        return self._fp.readline(size)

    def readlines(self, sizehint=-1):
        """
        Read as lines.

        :param sizehint: If the optional sizehint argument is present, instead of reading up to EOF,
                     whole lines totalling approximately sizehint bytes
                     (possibly after rounding up to an internal buffer size) are read.
        :return: lines
        :rtype: list
        """

        self._check_read()
        return self._fp.readlines(sizehint)

    def _check_write(self):
        if not self.opened:
            raise IOError('I/O operation on non-open resource')
        if self.mode == FileResource.Mode.READ:
            raise IOError('Resource not open for writing')

    def write(self, content):
        """
        Write content into the file resource

        :param content: content to write
        :return: None
        """
        self._check_write()
        return self._fp.write(content)

    def writelines(self, seq):
        """
        Write lines into the file resource.

        :param seq: lines
        :return: None
        """
        self._check_write()
        return self._fp.writelines(seq)

    def seek(self, pos, whence=compat.SEEK_SET):  # io.SEEK_SET
        """
        Seek to some place.

        :param pos: position to seek
        :param whence: if set to 2, will seek to the end
        :return: None
        """
        return self._fp.seek(pos, whence)

    def seekable(self):
        return self._fp.seekable()

    def tell(self):
        """
        Tell the current position

        :return: current position
        """
        return self._fp.tell()

    def truncate(self, size=None):
        """
        Truncate the file resource's size.

        :param size: If the optional size argument is present,
                     the file is truncated to (at most) that size.
                     The size defaults to the current position.
        :return: None
        """
        self._check_write()
        return self._fp.truncate(size)

    def flush(self):
        """
        Commit the change to ODPS if any change happens.
        Close will do this automatically.

        :return: None
        """
        return self._fp.flush()

    def close(self):
        """
        Close this file resource.

        :return: None
        """
        if self._fp is not None:
            self._fp.close()
            self._fp = None

    def __iter__(self):
        self._check_read()
        return self._fp.__iter__()

    def __next__(self):
        self._check_read()
        return next(self._fp)

    next = __next__

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def update(self, file_obj, temp=None):
        return self._parent.update(self, file_obj=file_obj, temp=temp)


@cache_parent
class JarResource(FileResource):
    """
    File resource representing for the .jar file.
    """

    def __init__(self, **kw):
        super(JarResource, self).__init__(**kw)
        self.type = Resource.Type.JAR


@cache_parent
class PyResource(FileResource):
    """
    File resource representing for the .py file.
    """

    def __init__(self, **kw):
        super(PyResource, self).__init__(**kw)
        self.type = Resource.Type.PY


@cache_parent
class ArchiveResource(FileResource):
    """
    File resource representing for the compressed file like .zip/.tgz/.tar.gz/.tar/jar
    """

    def __init__(self, **kw):
        super(ArchiveResource, self).__init__(**kw)
        self.type = Resource.Type.ARCHIVE


@cache_parent
class TableResource(Resource):
    """
    Take a table as a resource.
    """

    _TableSource = namedtuple("_TableSource", "project schema table partition")

    def __init__(self, **kw):
        project_name = kw.pop('project_name', None)
        schema_name = kw.pop('schema_name', None)
        table_name = kw.pop('table_name', None)
        partition_spec = kw.pop('partition', None)

        super(TableResource, self).__init__(**kw)

        self._init(
            table_project_name=project_name,
            table_schema_name=schema_name,
            table_name=table_name,
            partition=partition_spec,
            create=True,
        )

    def create(self, overwrite=False, **kw):
        if self.name is None or len(self.name.strip()) == 0:
            raise errors.ODPSError('Table Resource Name should not be empty.')

        method = self._client.post if not overwrite else self._client.put
        url = self.parent.resource() if not overwrite else self.resource()

        headers = {
            'Content-Type': 'text/plain',
            'x-odps-resource-type': self.type.value.lower(),
            'x-odps-resource-name': self.name,
            'x-odps-copy-table-source': self.source_table_name,
        }
        if self._getattr('comment') is not None:
            headers['x-odps-comment'] = self._getattr('comment')

        method(url, '', headers=headers, curr_schema=self._get_schema_name())

        if overwrite:
            del self.parent[self.name]
            return self.parent[self.name]
        return self

    def _init(self, create=False, table_project_name=None, table_schema_name=None, table_name=None, **kw):
        table_project_name = table_project_name or kw.get("project_name")
        if table_name is not None and '.' in table_name:
            parts = table_name.split('.')
            if len(parts) == 2:
                assert table_schema_name is None
                table_project_name, table_name = parts
            else:
                table_project_name, table_schema_name, table_name = parts

            table_project_name = table_project_name.strip()
            if table_schema_name is not None:
                table_schema_name = table_schema_name.strip()
            table_name = table_name.strip()

            if table_name.startswith("`") and table_name.endswith("`"):
                table_name = table_name[1:-1]

        try:
            if not create:
                table_source = self._get_table_source()
                old_table_project_name = table_source.project
                old_schema_name = table_source.schema
                old_table_name = table_source.table
                old_partition = table_source.partition
            else:
                old_table_project_name, old_schema_name, old_table_name, old_partition = [None] * 4
        except AttributeError:
            old_table_project_name, old_schema_name, old_table_name, old_partition = [None] * 4

        table_project_name = table_project_name or old_table_project_name or self.project.name
        table_schema_name = table_schema_name or old_schema_name or self._get_schema_name()
        table_name = table_name or old_table_name
        partition = kw.get('partition', old_partition)

        if table_name is not None:
            if table_schema_name:
                self.source_table_name = '.'.join((table_project_name, table_schema_name, table_name))
            else:
                self.source_table_name = '.'.join((table_project_name, table_name))

        if partition is not None:
            if not isinstance(partition, types.PartitionSpec):
                partition_spec = types.PartitionSpec(partition)
            else:
                partition_spec = partition
            self.source_table_name = (
                '%s partition(%s)' % (
                    self.source_table_name.split(' partition(')[0], partition_spec
                )
            )

    def _get_table_source(self):
        if self.source_table_name is None:
            raise AttributeError('source_table_name not defined.')

        splits = self.source_table_name.split(' partition(')
        if len(splits) < 2:
            partition = None
        else:
            partition = splits[1].split(')', 1)[0].strip()

        src = splits[0]
        if '.' not in src:
            raise ValueError('Malformed source table name: %s' % src)
        table_parts = src.split('.')
        if len(table_parts) == 2:
            schema_name = None
            project_name, table_name = table_parts
        else:
            project_name, schema_name, table_name = table_parts

        return TableResource._TableSource(
            project_name, schema_name, table_name, partition
        )

    def get_source_table(self):
        try:
            table_source = self._get_table_source()
        except AttributeError:
            return

        tables_parent = self.project.parent[table_source.project]
        if table_source.schema:
            tables_parent = tables_parent.schemas[table_source.schema]
        return tables_parent.tables[table_source.table]

    def get_source_table_partition(self):
        if self.source_table_name is None:
            return

        splits = self.source_table_name.split(' partition(')
        if len(splits) < 2:
            return

        partition = splits[1].split(')', 1)[0].strip()
        return types.PartitionSpec(partition)

    @property
    def table(self):
        """
        Get the table object.

        :return: source table
        :rtype: :class:`odps.models.Table`

        .. seealso:: :class:`odps.models.Table`
        """

        return self.get_source_table()

    @property
    def partition(self):
        """
        Get the source table partition.

        :return: the source table partition
        """

        pt = self.get_source_table_partition()
        if pt is None:
            return
        return self.get_source_table().get_partition(pt)

    def open_reader(self, **kwargs):
        """
        Open reader on the table resource
        """
        return self.get_source_table().open_reader(
            partition=self.get_source_table_partition(), **kwargs
        )

    def open_writer(self, **kwargs):
        """
        Open writer on the table resource
        """
        return self.get_source_table().open_writer(
            partition=self.get_source_table_partition(), **kwargs
        )

    def update(
        self, table_project_name=None, table_schema_name=None, table_name=None, *args, **kw
    ):
        """
        Update this resource.

        :param table_project_name: the source table's project
        :param table_name: the source table's name
        :param partition: the source table's partition
        :return: self
        """
        if len(args) > 0:
            kw['partition'] = args[0]
        self._init(
            table_project_name=table_project_name,
            table_schema_name=table_schema_name,
            table_name=table_name,
            **kw
        )
        resources = self.parent
        return resources.update(self)


@cache_parent
class VolumeResource(Resource):
    def create(self, overwrite=False, **kw):
        if self.name is None or len(self.name.strip()) == 0:
            raise errors.ODPSError('Volume Resource Name should not be empty.')

        method = self._client.post if not overwrite else self._client.put
        url = self.parent.resource() if not overwrite else self.resource()

        headers = {
            'Content-Type': 'text/plain',
            'x-odps-resource-type': self.type.value.lower(),
            'x-odps-resource-name': self.name,
            'x-odps-copy-file-source': self.volume_path,
        }
        if self._getattr('comment') is not None:
            headers['x-odps-comment'] = self._getattr('comment')

        method(url, '', headers=headers, curr_schema=self._get_schema_name())

        if overwrite:
            del self.parent[self.name]
            return self.parent[self.name]
        return self


@cache_parent
class VolumeFileResource(VolumeResource):
    """
    Volume resource represents for a volume archive
    """

    def __init__(self, **kw):
        okw = kw.copy()
        okw.pop('volume_file', None)
        super(VolumeFileResource, self).__init__(**okw)
        self.type = Resource.Type.VOLUMEFILE

    def create(self, overwrite=False, **kw):
        if 'volume_file' in kw:
            vf = kw.pop('volume_file')
            self.volume_path = vf.path
        return super(VolumeFileResource, self).create(overwrite, **kw)


@cache_parent
class VolumeArchiveResource(VolumeFileResource):
    """
    Volume archive resource represents for a volume archive
    """

    def __init__(self, **kw):
        super(VolumeArchiveResource, self).__init__(**kw)
        self.type = Resource.Type.VOLUMEARCHIVE



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

from ...df import DataFrame
from ...compat import six
from ...utils import underline_to_capitalized


class ColImportanceResult(dict):
    def __init__(self, stats_type, **kwargs):
        super(ColImportanceResult, self).__init__(**kwargs)
        self._titles = [(fn, underline_to_capitalized(fn)) for fn in stats_type._fields]

    def _repr_html_(self):
        sio = six.StringIO()
        sio.write('<table><tr><th>Column</th>')
        for _, ft in self._titles:
            sio.write('<th>{0}</th>'.format(ft))
        sio.write('</tr>')
        for col, stats in six.iteritems(self):
            sio.write('<tr><td>{0}</td>'.format(col))
            for fn, _ in self._titles:
                sio.write('<td>{0}</td>'.format(getattr(stats, fn)))
            sio.write('</tr>')
        sio.write('</table>')
        return sio.getvalue()


def get_rf_importance(expr, odps):
    stats_type = namedtuple('RFFeature', 'gini entropy')
    mapping = dict((row[0], stats_type(gini=row[1], entropy=row[2]))
                   for row in DataFrame(odps.get_table(expr.tables[0])).execute())
    return ColImportanceResult(stats_type, **mapping)


def get_gbdt_importance(expr, odps):
    stats_type = namedtuple('GBDTFeature', 'importance')
    mapping = dict((row[0], stats_type(importance=row[1]))
                   for row in DataFrame(odps.get_table(expr.tables[0])).execute())
    return ColImportanceResult(stats_type, **mapping)


def get_regression_importance(expr, odps):
    stats_type = namedtuple('RegressionFeature', 'weight importance')
    mapping = dict((row[0], stats_type(weight=row[1], importance=row[2]))
                   for row in DataFrame(odps.get_table(expr.tables[0])).execute())
    return ColImportanceResult(stats_type, **mapping)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pkgutil
import sys

try:
    from .datasource import read_coo
    from .datastore import write_coo
except ImportError as e:
    if "CI_MODE" in os.environ and pkgutil.find_loader("mars") is not None:
        if (
            sys.version_info[0] > 2
        ):  # todo remove this check when mars package under py27mu is fixed
            raise
    read_coo, write_coo = None, None

__all__ = ["read_coo", "write_coo"]
